/*
 *                       ibblob.c
 *
 *   - The module access to InterBase for the Python.
 *     ( The functions for Blob object )
 *
 *                               Copyright (C) 1997,1998 by RIOS Corporation
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * RIOS Corporation makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.  By use of this software the user agrees to
 * indemnify and hold harmless RIOS Corporation from any  claims or
 * liability for loss arising out of such use.
 */


/* Deallocate the blob object */

#ifdef __STDC__
static void	*ibblob_dealloc
(
    ibblobobject	*self
)
#else
static void	*ibblob_dealloc( self )
    ibblobobject	*self ;
#endif
{
	if( self->blob )
	    free( self->blob ) ;

	PyMem_DEL( self ) ;
	return ;
}



/* Get the attribute of blob object */

#ifdef __STDC__
static PyObject	*ibblob_getattr
(
    ibblobobject	*self,
    char		*attr_name
)
#else
static PyObject	*ibblob_getattr( self, attr_name )
    ibblobobject	*self ;
    char		*attr_name ;
#endif
{
	unsigned short	 value ;


	if( !strcmp( attr_name, BLOBATTR_DSIZE ) )
	{
	    return( PyInt_FromLong( self->size ) ) ;
	}
	else if( !strcmp( attr_name, BLOBATTR_SEGNO ) )
	{
	    memcpy( ( void * )&value,
		( void * )&self->desc.blob_desc_segment_size, sizeof(value) ) ;
	    return( PyInt_FromLong( ( long )value ) ) ;
	}
	else if( !strcmp( attr_name, BLOBATTR_SUBTP ) )
	{
	    return( PyInt_FromLong( ( long )self->desc.blob_desc_subtype ) ) ;
	}
	else if( !strcmp( attr_name, BLOBATTR_CHSET ) )
	{
	    return( PyInt_FromLong( ( long )self->desc.blob_desc_charset ) ) ;
	}
	else if( !strcmp( attr_name, BLOBATTR_TABLE ) )
	{
	    return( PyString_FromString( ( char * )self->desc.blob_desc_relation_name ) ) ;
	}
	else if( !strcmp( attr_name, BLOBATTR_FIELD ) )
	{
	    return( PyString_FromString( ( char * )self->desc.blob_desc_field_name ) ) ;
	}

	return( Py_FindMethod( ibblob_methods, ( PyObject * )self, attr_name ) ) ;
}



/* Print the blob object */

#ifdef __STDC__
static int	 ibblob_print
(
    ibblobobject	*self,
    FILE		*fp,
    int			 flags
)
#else
static int	 ibblob_print( self, fp, flags )
    ibblobobject	*self ;
    FILE		*fp ;
    int			 flags ;
#endif
{
	long		 count ;


	if( self->blob )
	{
	    if( self->desc.blob_desc_subtype == BLOB_SBTP_TEXT )
	    {
		for( count = 0 ; count < self->size ; count ++ )
		    fputc( ( int )self->blob[count], fp ) ;
	    }
	    else
		fprintf( fp, "Blob(subtype:%d,size:%d)",
			self->desc.blob_desc_subtype, self->size ) ;
	}
	else
	    fprintf( fp, "Blob(<no data>)" ) ;

	return( 0 ) ;
}



/* Get the blob data from the blob object */

#ifdef __STDC__
static PyObject	*ibblob_getdata
(
    ibblobobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibblob_getdata( self, args )
    ibblobobject	*self ;
    PyObject		*args ;
#endif
{
	/* Check the arguments */

	if( !PyArg_ParseTuple( args, "" ) )
	    return( NULL ) ;


	/* Create the String object */

	return( PyString_FromStringAndSize( self->blob, self->size ) ) ;
}



/* Set the blob data into the blob object */

#ifdef __STDC__
static PyObject	*ibblob_setdata
(
    ibblobobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibblob_setdata( self, args )
    ibblobobject	*self ;
    PyObject		*args ;
#endif
{
	PyObject	*blob ;
	char		*data ;
	long		 size ;


	/* Check the arguments */

	if( !PyArg_ParseTuple( args, "O", &blob ) )
	    return( NULL ) ;


	/* Get the blob data */

	data = get_blob_from_string( blob, &size ) ;
	if( !data )
	    return( NULL ) ;


	/* Set the blob data into blob object */

	if( self->blob )
	    free( self->blob ) ;
	self->blob = data ;
	self->size = size ;


	Py_INCREF( Py_None ) ;
	return( Py_None ) ;
}



/* List of methods defined in blob object */

static struct PyMethodDef ibblob_methods[] =
{
	{ "getdata",	( PyCFunction )ibblob_getdata,	METH_VARARGS,	NULL },
	{ "setdata",	( PyCFunction )ibblob_setdata,	METH_VARARGS,	NULL },
	{ NULL,		NULL }
};
