/*
 *                       ibcommon.c
 *
 *   - The module access to InterBase for the Python. ( Common functions )
 *
 *                               Copyright (C) 1997,1998 by RIOS Corporation
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * RIOS Corporation makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.  By use of this software the user agrees to
 * indemnify and hold harmless RIOS Corporation from any  claims or
 * liability for loss arising out of such use.
 */


/* Set the SQL code to the status object */

#ifdef __STDC__
static void	set_pyerror
(
    long	 sqlcode,
    char	*message
)
#else
static void	set_pyerror( sqlcode, message )
    long	 sqlcode ;
    char	*message ;
#endif
{
	PyObject	*tuple ;


	/* Create tuple object */

	if( ( tuple = PyTuple_New( 2 ) ) == NULL )
	    return ;


	/* Set the SQL code and error message */

	PyTuple_SetItem( tuple, 0, PyInt_FromLong( sqlcode ) ) ;
	PyTuple_SetItem( tuple, 1, PyString_FromString( message ) ) ;


	/* Set the tuple to the error object */

	PyErr_SetObject( ErrorObject, tuple ) ;


	return ;
}



/* Get the error message from the database */

#ifdef __STDC__
static void	set_dberror
(
    ISC_STATUS	 *status
)
#else
static void	set_dberror( status )
    ISC_STATUS	 *status ;
#endif
{
	ISC_STATUS	*vector ;
	int		 length ;


	vector = status ;
	isc_interprete( tempbuff_S, &vector ) ;
	strcat( tempbuff_S, "\n" ) ;
	length = strlen( tempbuff_S ) ;
	while( isc_interprete( &tempbuff_S[length], &vector ) )
	{
	    strcat( tempbuff_S, "\n" ) ;
	    length = strlen( tempbuff_S ) ;
	}

	set_pyerror( SQLCODE, tempbuff_S ) ;

	return ;
}



/* Allocate the XSQLDA structure */

#ifdef __STDC__
static XSQLDA	*alloc_sqlda
(
    long	 cols
)
#else
static XSQLDA	*alloc_sqlda( cols )
    long	 cols ;
#endif
{
	XSQLDA		*sqlda ;


	sqlda = ( XSQLDA * )malloc( XSQLDA_LENGTH( cols ) ) ;
	if( sqlda == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    return( NULL ) ;
	}
	sqlda->version = SQLDA_VERSION1 ;
	sqlda->sqln = cols ;
	sqlda->sqld = cols ;
	return( sqlda ) ;
}



/* Reallocate buffer */

#ifdef __STDC__
static char	*realloc_buffer
(
    char	*ptr,
    long	 size
)
#else
static char	*realloc_buffer( ptr, size )
    char	*ptr ;
    long	 size ;
#endif
{
	char	*new_p ;


	if( ptr )
	    new_p = ( char * )realloc( ( void * )ptr, size + 1 ) ;
	else
	    new_p = ( char * )malloc( size + 1 ) ;

	if( new_p == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    return( NULL ) ;
	}

	return( new_p ) ;
}



/* Get the keyword parameters */

#ifdef __STDC__
static int	get_keywordargs
(
    PyObject	*kwd,
    keywd_args	*list
)
#else
static int	get_keywordargs( kwd, list )
    PyObject	*kwd ;
    keywd_args	*list ;
#endif
{
	int		 d_count ;
	PyObject	*d_key ;
	PyObject	*d_value ;
	char		*keyword ;
	char		*value ;
	int		 count ;


	/* Check the keyword dictionary */

	if( !kwd )
	    return( STATUS_NORMAL ) ;

	if( !PyDict_Check( kwd ) )
	{
	    set_pyerror( 0, ERR_ILGLKWDIC ) ;
	    return( STATUS_ABNORMAL ) ;
	}


	/* Get the keyword parameters */

	d_count = 0 ;
	while( PyDict_Next( kwd, &d_count, &d_key, &d_value ) )
	{
	    if( !PyString_Check( d_key ) || !PyString_Check( d_value ) )
	    {
		set_pyerror( 0, ERR_NOTSTRING ) ;
		return( STATUS_ABNORMAL ) ;
	    }
	    keyword = PyString_AS_STRING( ( PyStringObject * )d_key ) ;
	    value = PyString_AS_STRING( ( PyStringObject * )d_value ) ;

	    for( count = 0 ; list[count].keyword ; count ++ )
	    {
		if( !strcmp( keyword, list[count].keyword ) )
		{
		    if( list[count].value )
		    {
			set_pyerror( 0, ERR_DUPKWDARG ) ;
			return( STATUS_ABNORMAL ) ;
		    }
		    list[count].value = value ;
		    break ;
		}
	    }

	    if( !list[count].keyword )
	    {
		sprintf( tempbuff_S, ERR_ILGLKWDPM, keyword ) ;
		set_pyerror( 0, tempbuff_S ) ;
		return( STATUS_ABNORMAL ) ;
	    }
	}

	return( STATUS_NORMAL ) ;
}



/* Calculate the size of XSQLDA parameter's bufffer */

#ifdef __STDC__
static long	calculate_buffsize
(
    XSQLDA	*sqlda
)
#else
static long	calculate_buffsize( sqlda )
    XSQLDA	*sqlda ;
#endif
{
	long		 cols ;
	XSQLVAR		*var ;
	short		 dtp ;
	long		 offset ;
	long		 alignment ;
	long		 length ;
	long		 count ;



	cols = sqlda->sqld ;
	for( var = sqlda->sqlvar, offset = 0, count = 0 ;
				count < cols ; var ++, count ++ )
	{
	    length = alignment = var->sqllen ;
	    dtp = var->sqltype & ~1 ;

	    if( dtp == SQL_TEXT )
		alignment = 1 ;
	    else if( dtp == SQL_VARYING )
	    {
		length += sizeof( short ) ;
		alignment = sizeof( short ) ;
	    }

	    offset = ALIGN( offset, alignment ) ;
	    offset += length ;
	    offset = ALIGN( offset, sizeof( short ) ) ;
	    offset += sizeof( short ) ;
	}

	return( offset ) ;
}



/* Free the description */

#ifdef __STDC__
static void	free_description
(
    PyObject	*desc
)
#else
static void	free_description( desc )
    PyObject	*desc ;
#endif
{
	Py_DECREF( desc ) ;

	return ;
}



/* Close all cursors */

#ifdef __STDC__
static void	close_allcursors
(
    ibcursobject	*first_p
)
#else
static void	close_allcursors( first_p )
    ibcursobject	*first_p ;
#endif
{
	ibcursobject	*curs_p ;
	ibcursobject	*next_p ;


	/* Close all cursors */

	curs_p = first_p ; 
	while( curs_p ) 
	{
	    next_p = curs_p->next_p ;
	    if( curs_p->open_f )
	    {
		isc_dsql_free_statement( isc_status,
					&curs_p->STMT_hdl, DSQL_close ) ;
		curs_p->open_f = IBOBJ_INIT ;
	    }
	    if( curs_p->describe )
	    {
		Py_DECREF( curs_p->describe ) ;
		curs_p->describe = NULL ;
	    }
	    curs_p = next_p ;
	}

	return ;
}



/* Get the database No. by the alias name */

#ifdef __STDC__
static long	get_database_number
(
    long		 DB_num,
    DBinformation	*DB_info,
    char		*alias
)
#else
static long	get_database_number( DB_num, DB_info, alias )
    long		 DB_num ;
    DBinformation	*DB_info ;
    char		*alias ;
#endif
{
	DBinformation	*curr_p ;
	long		 count ;
	long		 DB_no ;


	if( !alias || ( int )strlen( alias ) == 0 )
	    return( 0 ) ;

	DB_no = STATUS_ABNORMAL ;
	for( count = 0 ; count < DB_num ; count ++ )
	{
	    curr_p = &DB_info[count] ;
	    if( curr_p->db_alias && !strcmp( alias, curr_p->db_alias ) )
	    {
		DB_no = count ;
		break ;
	    }
	}

	return( DB_no ) ;
}



/* Connect to the database */

#ifdef __STDC__
static int	connect_database
(
    DBinformation	*DB_info,
    char		*dbname,
    char		*charset,
    char		*usname,
    char		*passwd,
    char		*alias
)
#else
static int	connect_database( DB_info, dbname,
				charset, usname, passwd, alias )
    DBinformation	*DB_info ;
    char		*dbname ;
    char		*charset ;
    char		*usname ;
    char		*passwd ;
    char		*alias ;
#endif
{
	isc_db_handle	 DB_hdl ;
	short		 dpb_len ;
	char		 dpb ;
	char		*dpb_p ;
	char		*ptr ;


	/* Initialize */

	DB_info->db_alias = NULL ;
	DB_info->DB_hdl = NULL ;
	DB_info->tpb_len = 0 ;
	DB_info->tpb_ptr = NULL ;
	DB_info->charset = isc_interp_eng_ascii ;
	DB_info->username = NULL ;
	DB_info->password = NULL ;

	SETNULL( charset ) ;
	SETNULL( usname ) ;
	SETNULL( passwd ) ;
	SETNULL( alias ) ;


	/* Set the DPB */

	dpb_len = 1 ;
	dpb = isc_dpb_version1 ;
	dpb_p = &dpb ;
	isc_expand_dpb( &dpb_p,		&dpb_len,
		isc_dpb_user_name,	usname,
		isc_dpb_password,	passwd,
		isc_dpb_lc_ctype,	charset,
		0 ) ;


	/* Connect to the database */

	DB_hdl = 0 ;
	isc_attach_database( isc_status, 0, dbname, &DB_hdl, dpb_len, dpb_p ) ;
	if( dpb_p != &dpb )	isc_free( dpb_p ) ;
	CHECK_IBERR2( isc_status, STATUS_ABNORMAL ) ;

	DB_info->DB_hdl = DB_hdl ;

	if( alias )
	{
	    DB_info->db_alias = strdup( alias ) ;
	    if( DB_info->db_alias == NULL )
	    {
		set_pyerror( 0, ERR_ALLOCMEM ) ;
		goto Error_Proc ;
	    }
	}

	if( charset )
	{
	    strncpy( tempbuff_S, charset, META_NAME_MAX ) ;
	    tempbuff_S[META_NAME_MAX] = '\0' ;
	    for( ptr = tempbuff_S ; *ptr ; ptr ++ )
		*ptr = UPPER( *ptr ) ;
	    if( !strcmp( tempbuff_S, IBCHARSET_EUC ) )
		DB_info->charset = isc_interp_jpn_euc ;
	    else if( !strcmp( tempbuff_S, IBCHARSET_SJIS ) )
		DB_info->charset = isc_interp_jpn_sjis ;
	}

	if( usname )
	{
	    DB_info->username = strdup( usname ) ;
	    if( DB_info->username == NULL )
	    {
		set_pyerror( 0, ERR_ALLOCMEM ) ;
		goto Error_Proc ;
	    }
	}

	if( passwd )
	{
	    DB_info->password = strdup( passwd ) ;
	    if( DB_info->password == NULL )
	    {
		set_pyerror( 0, ERR_ALLOCMEM ) ;
		goto Error_Proc ;
	    }
	}


	return( STATUS_NORMAL ) ;


Error_Proc:
	isc_detach_database( isc_status, &DB_hdl ) ;
	FREE( DB_info->db_alias ) ;
	FREE( DB_info->username ) ;
	FREE( DB_info->password ) ;
	return( STATUS_ABNORMAL ) ;
}



/* Disconnect from the database */

#ifdef __STDC__
static void	disconnect_database
(
    long		 DB_num,
    DBinformation	*DB_info
)
#else
static void	disconnect_database( DB_num, DB_info )
    long		 DB_num ;
    DBinformation	*DB_info ;
#endif
{
	DBinformation	*curr_p ;
	long		 count ;


	for( count = 0 ; count < DB_num ; count ++ )
	{
	    curr_p = &DB_info[count] ;
	    if( curr_p->DB_hdl )
		isc_detach_database( isc_status, &curr_p->DB_hdl ) ;
	    FREE( curr_p->db_alias ) ;
	    FREE( curr_p->tpb_ptr ) ;
	    FREE( curr_p->username ) ;
	    FREE( curr_p->password ) ;
	}

	return ;
}



/* Disconnect from the database and clear all DB informations */

#ifdef __STDC__
static void	clear_all_database
(
    ibconnobject	*self
)
#else
static void	clear_all_database( self )
    ibconnobject	*self ;
#endif
{
	ibcursobject	*curs_p ;


	/* Close the cursors */

	close_allcursors( self->first_p ) ;


	/* Unlink the cursor objects */

	for( curs_p = self->first_p ; curs_p ; curs_p = curs_p->next_p )
	    curs_p->conn_p = NULL ;

	self->first_p = NULL ;


	/* Rollback the transaction and disconnect database */

	if( self->TR_hdl )
	{
	    isc_rollback_transaction( isc_status, &self->TR_hdl ) ;
	}

	disconnect_database( self->DB_num, self->DB_info ) ;
	self->DB_num = 0 ;
	FREE( self->DB_info ) ;
	self->DB_info = NULL ;


	return ;
}



/* Create a Connection object */

#ifdef __STDC__
static PyObject	*create_connection
(
    long		 DB_num,
    DBinformation	*DB_info
)
#else
static PyObject	*create_connection( DB_num, DB_info )
    long		 DB_num ;
    DBinformation	*DB_info ;
#endif
{
	ibconnobject	*newobj ;



	/* Create the connection object */

	newobj = PyObject_NEW( ibconnobject, &IBconntype ) ;
	if( newobj == NULL )
	    goto Error_Proc ;

	newobj->DB_num = DB_num ;
	newobj->DB_info = DB_info ;
	newobj->TR_hdl = NULL ;
	newobj->teb_size = 0 ;
	newobj->teb_vec = NULL ;
	newobj->first_p = NULL ;
	newobj->open_f = IBOBJ_INIT ;


	/* Allocate buffers */

	newobj->i_sqlda = alloc_sqlda( SQLDA_COLSINIT ) ;
	if( newobj->i_sqlda == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    goto Error_Proc ;
	}

	newobj->o_sqlda = alloc_sqlda( SQLDA_COLSINIT ) ;
	if( newobj->o_sqlda == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    goto Error_Proc ;
	}

	newobj->results = realloc_buffer( NULL, RESULT_BUFINIT ) ;
	if( newobj->results == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    goto Error_Proc ;
	}
	newobj->ressize = RESULT_BUFINIT ;

	newobj->paramts = realloc_buffer( NULL, PARAMT_BUFINIT ) ;
	if( newobj->paramts == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    goto Error_Proc ;
	}
	newobj->prmsize = PARAMT_BUFINIT ;


	newobj->open_f = IBOBJ_OPEN ;

	return( ( PyObject * )newobj ) ;


Error_Proc:
	if( newobj )
	{
	    newobj->DB_info = NULL ;
	    Py_XDECREF( newobj ) ;
	}
	return( NULL ) ;
}



/* Check validation of the specified date */

#ifdef __STDC__
static int	check_validdate
(
    struct tm		*tms
)
#else
static int	check_validdate( tms )
    struct tm		*tms ;
#endif
{
	ISC_QUAD	 ibdt ;
	struct tm	 tms2 ;



	/* Check the date */

	if( tms->tm_hour < 0	|| tms->tm_hour > 23
	||  tms->tm_min < 0	|| tms->tm_min > 59
	||  tms->tm_sec < 0	|| tms->tm_sec > 59	)
	{
	    set_pyerror( 0, ERR_ILLGLDATE ) ;
	    return( STATUS_ABNORMAL ) ;
	}

	isc_encode_date( tms, &ibdt ) ;
	isc_decode_date( &ibdt, &tms2 ) ;

	if( tms->tm_year != tms2.tm_year
	||  tms->tm_mon	 != tms2.tm_mon
	||  tms->tm_mday != tms2.tm_mday
	||  tms->tm_hour != tms2.tm_hour
	||  tms->tm_min	 != tms2.tm_min
	||  tms->tm_sec	 != tms2.tm_sec	)
	{
	    set_pyerror( 0, ERR_ILLGLDATE ) ;
	    return( STATUS_ABNORMAL ) ;
	}

	*tms = tms2 ;
	tms->tm_isdst = -1 ;

	return( STATUS_NORMAL ) ;
}



/* Create a tuple from the tm structure */

#ifdef __STDC__
static PyObject	*date_tuple_from_tm
(
    struct tm		*tms
)
#else
static PyObject	*date_tuple_from_tm( tms )
    struct tm		*tms ;
#endif
{
	PyObject	*tuple ;


	/* Create the result tuple object */

	if( ( tuple = PyTuple_New( DATEOBJ_NUMBER ) ) == NULL )
	    return( NULL ) ;


	/* Set the date items */

	PyTuple_SetItem( tuple, DATEITM_YEAR,
			  PyInt_FromLong( tms->tm_year + 1900 ) ) ;
	PyTuple_SetItem( tuple, DATEITM_MONTH,
			  PyInt_FromLong( tms->tm_mon + 1 ) ) ;
	PyTuple_SetItem( tuple, DATEITM_DAY,
			  PyInt_FromLong( tms->tm_mday ) ) ;
	PyTuple_SetItem( tuple, DATEITM_HOUR,
			  PyInt_FromLong( tms->tm_hour ) ) ;
	PyTuple_SetItem( tuple, DATEITM_MIN,
			  PyInt_FromLong( tms->tm_min ) ) ;
	PyTuple_SetItem( tuple, DATEITM_SEC,
			  PyInt_FromLong( tms->tm_sec ) ) ;

	return( tuple ) ;
}



/* Set values in tuple or list into tm structure  */

#ifdef __STDC__
static int	 tm_from_date_tuple
(
    PyObject		*params,
    struct tm		*tms
)
#else
static int	 tm_from_date_tuple( params, tms )
    PyObject		*params ;
    struct tm		*tms ;
#endif
{
	PyObject	*( * func_getitem )() ;
	long		 prm_no ;
	long		 count ;
	PyObject	*obj ;
	long		 value ;


	/* Get the item number */

	if( PyTuple_Check( params ) )
	{
	    prm_no = PyTuple_Size( params ) ;
	    func_getitem = PyTuple_GetItem ;
	}
	else if( PyList_Check( params ) )
	{
	    prm_no = PyList_Size( params ) ;
	    func_getitem = PyList_GetItem ;
	}
	else
	{
	    set_pyerror( 0, ERR_DATEOTYPE ) ;
	    return( STATUS_ABNORMAL ) ;
	}

	if( prm_no <= 0 || prm_no > DATEOBJ_NUMBER )
	{
	    set_pyerror( 0, ERR_NUMITEMDT ) ;
	    return( STATUS_ABNORMAL ) ;
	}


	/* Initialize */

	tms->tm_mon = 0 ;
	tms->tm_mday = 1 ;
	tms->tm_hour = 0 ;
	tms->tm_min = 0 ;
	tms->tm_sec = 0 ;
	tms->tm_isdst = -1 ;


	/* Set the item values */

	for( count = 0 ; count < prm_no ; count ++ )
	{
	    obj = func_getitem( params, count ) ;
	    if( !PyInt_Check( obj ) )
	    {
		set_pyerror( 0, ERR_DATEOTYPE ) ;
		return( STATUS_ABNORMAL ) ;
	    }

	    value = PyInt_AS_LONG( ( PyIntObject * )obj ) ;

	    switch( count )
	    {
		case DATEITM_YEAR :
		    tms->tm_year = value - 1900 ;
		    break ;
		case DATEITM_MONTH :
		    tms->tm_mon = value - 1 ;
		    break ;
		case DATEITM_DAY :
		    tms->tm_mday = value ;
		    break ;
		case DATEITM_HOUR :
		    tms->tm_hour = value ;
		    break ;
		case DATEITM_MIN :
		    tms->tm_min = value ;
		    break ;
		case DATEITM_SEC :
		    tms->tm_sec = value ;
		    break ;
	    }
	}


	/* Check the date */

	if( check_validdate( tms ) )
	    return( STATUS_ABNORMAL ) ;


	return( STATUS_NORMAL ) ;
}



/* Create a new date object from the tm structure  */

#ifdef __STDC__
static ibdateobject	*new_date_object
(
    struct tm		*tms
)
#else
static ibdateobject	*new_date_object( tms )
    struct tm		*tms ;
#endif
{
	ibdateobject	*newobj ;


	/* Create the date object */

	newobj = PyObject_NEW( ibdateobject, &IBdatetype ) ;
	if( newobj == NULL )
	    return( NULL ) ;


	/* Set the new values */

	newobj->dtime = *tms ;
	newobj->dtime.tm_isdst = -1 ;


	return( newobj ) ;
}



/* Create a blob from the String object */

#ifdef __STDC__
static char	*get_blob_from_string
(
    PyObject		*blob,
    long		*size
)
#else
static char	*get_blob_from_string( blob, size )
    PyObject		*blob ;
    long		*size ;
#endif
{
	long		 length ;
	char		*b_ptr ;


	/* Get the blob data */

	if( !PyString_Check( blob ) )
	{
	    set_pyerror( 0, ERR_BLOBDATTP ) ;
	    return( NULL ) ;
	}

	length = PyString_Size( blob ) ;
	b_ptr = ( char * )malloc( length + 1 ) ;
	if( b_ptr == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    return( NULL ) ;
	}

	memcpy( ( void * )b_ptr, ( void * )
		PyString_AS_STRING( ( PyStringObject * )blob ), length ) ;
	b_ptr[length] = '\0' ;

	*size = length ;

	return( b_ptr ) ;
}



/* Create a new blob object */

#ifdef __STDC__
static ibblobobject	*new_blob_object
(
    long		 size,
    char		*data,
    ISC_BLOB_DESC	*desc
)
#else
static ibblobobject	*new_blob_object( size, data, desc )
    long		 size ;
    char		*data ;
    ISC_BLOB_DESC	*desc ;
#endif
{
	ibblobobject	*newobj ;


	/* Create the blob object */

	newobj = PyObject_NEW( ibblobobject, &IBblobtype ) ;
	if( newobj == NULL )
	    return( NULL ) ;

	newobj->size = size ;
	newobj->blob = data ;
	newobj->desc = *desc ;


	return( newobj ) ;
}



/* Create a new blob object from the string object */

#ifdef __STDC__
static ibblobobject	*new_blob_from_string
(
    ibconnobject	*conn,
    long		 DB_no,
    char		*table,
    char		*field,
    PyObject		*blob
)
#else
static ibblobobject	*new_blob_from_string( conn, DB_no, table, field, blob )
    ibconnobject	*conn ;
    long		 DB_no ;
    char		*table ;
    char		*field ;
    PyObject		*blob ;
#endif
{
	char		*b_ptr = NULL ;
	long		 size = 0 ;
	ISC_BLOB_DESC	 b_desc ;
	int		 length ;
	char		*ptr ;
	DBinformation	*DB_info ;
	isc_tr_handle	 TR_hdl ;



	/* Check the arguments */

	if( table == NULL || field == NULL
	|| ( int )strlen( table ) == 0 || ( int )strlen( field ) == 0 )
	{
	    set_pyerror( 0, ERR_BLOBOBJCT ) ;
	    return( NULL ) ;
	}


	/* Get the blob data */

	if( blob )
	{
	    b_ptr = get_blob_from_string( blob, &size ) ;
	    if( !b_ptr )
		return( NULL ) ;
	}


	/* Start the transaction */

	DB_info = &conn->DB_info[DB_no] ;
	TR_hdl = NULL ;
	isc_start_transaction( isc_status, &TR_hdl, ( short )1,
		&DB_info->DB_hdl, ( short )sizeof( isc_tpb_0 ), isc_tpb_0 ) ;
	CHECK_IBERR2( isc_status, NULL ) ;


	/* Get the blob description */

	length = sizeof( b_desc.blob_desc_relation_name ) - 1 ;
	strncpy( ( char * )b_desc.blob_desc_relation_name, table, length ) ;
	b_desc.blob_desc_relation_name[length] = '\0' ;
	for( ptr = ( char * )b_desc.blob_desc_relation_name ; *ptr ; ptr ++ )
	    *ptr = UPPER( *ptr ) ;

	length = sizeof( b_desc.blob_desc_field_name ) - 1 ;
	strncpy( ( char * )b_desc.blob_desc_field_name, field, length ) ;
	b_desc.blob_desc_field_name[length] = '\0' ;
	for( ptr = ( char * )b_desc.blob_desc_field_name ; *ptr ; ptr ++ )
	    *ptr = UPPER( *ptr ) ;

	isc_blob_lookup_desc( isc_status, &DB_info->DB_hdl,
			&TR_hdl, b_desc.blob_desc_relation_name,
			b_desc.blob_desc_field_name, &b_desc, NULL ) ;
	SQLCODE = isc_sqlcode( isc_status ) ;
	if( SQLCODE )
	{
	    set_dberror( isc_status ) ;
	    isc_rollback_transaction( isc_status, &TR_hdl ) ;
	    free( b_ptr ) ;
	    return( NULL ) ;
	}
	isc_rollback_transaction( isc_status, &TR_hdl ) ;


	/* Create the blob object */

	return( new_blob_object( size, b_ptr, &b_desc ) ) ;
}



/* Retrieve the blob data from database */

#ifdef __STDC__
static int	retrieve_blob
(
    ibcursobject	 *curs,
    XSQLVAR		 *var,
    long		 *src_len,
    char		**src_buf,
    ISC_BLOB_DESC	 *from_desc
)
#else
static int	retrieve_blob( curs, var, src_len, src_buf, from_desc )
    ibcursobject	 *curs ;
    XSQLVAR		 *var ;
    long		 *src_len ;
    char		**src_buf ;
    ISC_BLOB_DESC	 *from_desc ;
#endif
{
	ibconnobject	*conn = curs->conn_p ;
	DBinformation	*DB_info ;
	long		 status ;
	isc_blob_handle	 blob_hdl = NULL ;
	long		 isc_dummy[20] ;
	char		 blob_item ;
	char		 blob_rslt[BLOBINFO_RSLT] ;
	char		*ptr ;
	char		 item ;
	short		 item_len ;
	long		 length ;
	char		*blob_buff = NULL ;
	long		 read_len ;
	long		 buff_len ;
	unsigned short	 seg_len ;
	ISC_BLOB_DESC	 to_desc ;
	unsigned char	 bpb_buffer[64] ;
	char		*bpb ;
	unsigned short	 bpb_length ;



	/* Create blob parameter block */

	DB_info = &conn->DB_info[curs->DB_no] ;
	bpb = NULL ;
	status = isc_blob_lookup_desc( isc_status, &DB_info->DB_hdl,
			&conn->TR_hdl, ( unsigned char * )var->relname,
			( unsigned char * )var->sqlname, from_desc, NULL ) ;

	if( var->sqlsubtype == BLOB_SBTP_TEXT && !status )
	{
	    isc_blob_default_desc( &to_desc,
					( unsigned char * )var->relname,
					( unsigned char * )var->sqlname ) ;
	    if( !isc_blob_gen_bpb( isc_status, &to_desc, from_desc,
			sizeof( bpb_buffer ), bpb_buffer, &bpb_length ) )
		bpb = ( char * )bpb_buffer ;
	}
	if( !bpb )	bpb_length = 0 ;


	/* Open blob */

	if( isc_open_blob2( isc_status, &DB_info->DB_hdl, &conn->TR_hdl,
		&blob_hdl, ( ISC_QUAD * )var->sqldata, bpb_length, bpb ) )
	    goto DBErr_Proc ;


	/* Get the blob size */

	blob_item = isc_info_blob_total_length ;
	if( isc_blob_info( isc_status, &blob_hdl, 1, &blob_item,
					BLOBINFO_RSLT, blob_rslt ) )
	    goto DBErr_Proc ;

	length = 0 ;
	for( ptr = blob_rslt ; length == 0 && *ptr != isc_info_end ; )
	{
	    item = *ptr++ ;
	    item_len = ( short )isc_vax_integer( ptr, 2 ) ;
	    ptr += 2 ;
	    switch( item )
	    {
		case isc_info_blob_total_length :
		    length = isc_vax_integer( ptr, item_len ) ;
		    break ;
		default :
		    break ;
	    }
	    ptr += item_len ;
	}


	/* Allocate memory */

	blob_buff = ( char * )malloc( length + 1 ) ;
	if( blob_buff == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    goto Error_Proc ;
	}


	/* Retrieve blob segments */

	buff_len = length ;
	read_len = 0 ;
	for( ;; )
	{
	    /* Get the segment */

	    if( isc_get_segment( isc_status, &blob_hdl, &seg_len,
			( unsigned short )
			( buff_len > USHORT_MAX ? USHORT_MAX : buff_len ),
			&blob_buff[read_len] ) )
	    {
		break ;
	    }

	    read_len += seg_len ;
	    buff_len -= seg_len ;
	}
	if( isc_status[1] != isc_segstr_eof )
	    goto DBErr_Proc ;


	/* Close blob */

	if( isc_close_blob( isc_status, &blob_hdl ) )
	    goto Error_Proc ;


	/* Set the results */

	blob_buff[read_len] = '\0' ;
	*src_len = read_len ;
	*src_buf = blob_buff ;

	return( STATUS_NORMAL ) ;


DBErr_Proc:
	SQLCODE = isc_sqlcode( isc_status ) ;
	set_dberror( isc_status ) ;

Error_Proc:
	if( blob_hdl )
	    isc_close_blob( isc_dummy, &blob_hdl ) ;
	if( blob_buff )
	    free( blob_buff ) ;
	return( STATUS_ABNORMAL ) ;
}



/* Insert the blob data into database */

#ifdef __STDC__
static int	create_blob
(
    ibcursobject	*curs,
    int			 text_f,
    ISC_BLOB_DESC	*to_desc,
    long		 src_len,
    char		*src_buf,
    ISC_QUAD		*blobid
)
#else
static int	create_blob( curs, text_f, to_desc, src_len, src_buf, blobid )
    ibcursobject	*curs ;
    int			 text_f ;
    ISC_BLOB_DESC	*to_desc ;
    long		 src_len ;
    char		*src_buf ;
    ISC_QUAD		*blobid ;
#endif
{
	ibconnobject		*conn = curs->conn_p ;
	DBinformation		*DB_info ;
	char			*src_ptr ;
	unsigned char		*ptr ;
	int			 length ;
	isc_blob_handle		 blob_hdl = NULL ;
	ISC_QUAD		 blob_id ;
	unsigned short		 seg_len ;
	ISC_BLOB_DESC		 from_desc ;
	unsigned char		 bpb_buffer[64] ;
	char			*bpb ;
	unsigned short		 bpb_length ;
	long			 cut_seg ;



	/* create blob parameter block */

	DB_info = &conn->DB_info[curs->DB_no] ;
	bpb = NULL ;
	if( text_f && to_desc )
	{
	    isc_blob_default_desc( &from_desc,
				to_desc->blob_desc_relation_name,
				to_desc->blob_desc_field_name ) ;
	    if( isc_blob_gen_bpb( isc_status, to_desc, &from_desc,
			sizeof( bpb_buffer ), bpb_buffer, &bpb_length ) )
		goto Error_Proc ;

	    bpb = ( char * )bpb_buffer ;
	}

	if( !bpb )	bpb_length = 0 ;


	/* create blob handle */

	if( isc_create_blob2( isc_status, &DB_info->DB_hdl,
			&conn->TR_hdl, &blob_hdl, &blob_id, bpb_length, bpb ) )
	    goto Error_Proc ;


	/* Check the character set */

	cut_seg = 0 ;
	if( text_f && to_desc
	&& ( DB_info->charset == isc_interp_jpn_euc
	  || DB_info->charset == isc_interp_jpn_sjis ) )
	{
	    cut_seg = to_desc->blob_desc_segment_size ;
	}


	/* insert blob segments */

	length = src_len ;
	src_ptr = src_buf ;
	while( length > 0 )
	{
	    if( cut_seg )
	    {
		if( length > cut_seg )
		{
		    if( DB_info->charset == isc_interp_jpn_sjis )    /* SJIS */
		    {
			ptr = ( unsigned char * )src_ptr ;
			seg_len = 0 ;
			while( ( int )seg_len < cut_seg )
			{
			    if( ( *ptr >= 0x80 && *ptr <= 0x9f )
			    ||	( *ptr >= 0xe0 && *ptr <= 0xef ) )
			    {
				if( ( int )( seg_len + 2 ) > cut_seg )
				    break ;
				else
				{
				    seg_len += 2 ;
				    ptr += 2 ;
				}
			    }
			    else
			    {
				seg_len ++ ;
				ptr ++ ;
			    }
			}
		    }
		    else					/* EUC */
		    {
			ptr = ( unsigned char * )src_ptr ;
			seg_len = 0 ;
			while( ( int )seg_len < cut_seg )
			{
			    if( *ptr & 0x80 )
			    {
				if( ( int )( seg_len + 2 ) > cut_seg )
				    break ;
				else
				{
				    seg_len += 2 ;
				    ptr += 2 ;
				}
			    }
			    else
			    {
				seg_len ++ ;
				ptr ++ ;
			    }
			}
		    }
		}
		else
		{
		    seg_len = length ;
		}
	    }
	    else
	    {
		if( length > USHORT_MAX )
		    seg_len = USHORT_MAX ;
		else
		    seg_len = length ;
	    }

	    if( isc_put_segment( isc_status, &blob_hdl, seg_len, src_ptr ) )
		goto Error_Proc ;

	    length -= seg_len ;
	    src_ptr += seg_len ;
	}

	if( isc_close_blob( isc_status, &blob_hdl ) )
	    goto Error_Proc ;

	*blobid = blob_id ;

	return( STATUS_NORMAL ) ;


Error_Proc:
	SQLCODE = isc_sqlcode( isc_status ) ;
	set_dberror( isc_status ) ;

	if( blob_hdl )
	    isc_close_blob( isc_status, &blob_hdl ) ;

	return( STATUS_ABNORMAL ) ;
}



/* Retrieve the array data from database */

#ifdef __STDC__
static int	retrieve_array
(
    ibcursobject	 *curs,
    XSQLVAR		 *var,
    long		 *src_len,
    char		**src_buf
)
#else
static int	retrieve_array( curs, var, src_len, src_buf )
    ibcursobject	 *curs ;
    XSQLVAR		 *var ;
    long		 *src_len ;
    char		**src_buf ;
#endif
{
	ibconnobject	*conn = curs->conn_p ;
	DBinformation	*DB_info ;
	ISC_ARRAY_DESC	 desc ;
	char		*array_buf = NULL ;
	long		 buff_len ;
	int		 count ;
	long		 length ;
	long		 sum_dim ;

/**********************************
 * This function is not used now! *
 **********************************/

	/* get the array description */

	DB_info = &conn->DB_info[curs->DB_no] ;
	if( isc_array_lookup_bounds( isc_status, &DB_info->DB_hdl,
			&conn->TR_hdl, var->relname, var->sqlname, &desc ) )
	    goto Error_Proc ;

	if( desc.array_desc_dtype == blr_text
	||  desc.array_desc_dtype == blr_text2
	||  desc.array_desc_dtype == blr_varying
	||  desc.array_desc_dtype == blr_varying2 )
	{
	    length = desc.array_desc_length + 1 ;
	    desc.array_desc_dtype = blr_cstring ;
	}
	else if( desc.array_desc_dtype == blr_cstring
	||	 desc.array_desc_dtype == blr_cstring2 )
	{
	    length = desc.array_desc_length + 1 ;
	}
	else
	    length = desc.array_desc_length ;


	/* Allocate memory */

	sum_dim = 1 ;
	for( count = 0 ; count < desc.array_desc_dimensions ; count ++ )
	{
	    sum_dim *= desc.array_desc_bounds[count].array_bound_upper
			- desc.array_desc_bounds[count].array_bound_lower + 1 ;
	}

	buff_len = sum_dim * length ;
	array_buf = ( char * )malloc( buff_len ) ;
	if( array_buf == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    goto Error_Proc ;
	}


	/* get the array data */

	if( isc_array_get_slice( isc_status, &DB_info->DB_hdl, &conn->TR_hdl,
					( ISC_QUAD * )var->sqldata, &desc,
					( void * )array_buf, &buff_len ) )
	    goto Error_Proc ;


	/* Set the results */

	*src_len = buff_len ;
	*src_buf = array_buf ;


	return( STATUS_NORMAL ) ;


Error_Proc:
	SQLCODE = isc_sqlcode( isc_status ) ;
	set_dberror( isc_status ) ;
	if( array_buf )
	    free( array_buf ) ;
	return( STATUS_ABNORMAL ) ;
}



/* Insert the array data into database */

#ifdef __STDC__
static int	create_array
(
    ibcursobject	*curs,
    XSQLVAR		*var,
    long		 src_len,
    char		*src_buf
)
#else
static int	create_array( curs, var, src_len, src_buf )
    ibcursobject	*curs ;
    XSQLVAR		*var ;
    long		 src_len ;
    char		*src_buf ;
#endif
{
	ibconnobject		*conn = curs->conn_p ;
	DBinformation		*DB_info ;
	ISC_ARRAY_DESC		 desc ;

/**********************************
 * This function is not used now! *
 **********************************/

	/* get the array description */

	DB_info = &conn->DB_info[curs->DB_no] ;
	if( isc_array_lookup_bounds( isc_status, &DB_info->DB_hdl,
			&conn->TR_hdl, var->relname, var->sqlname, &desc ) )
			/************************************************
			 *  The variable var->relname and var->sqlname	*
			 *  is not set when execute INSERT or UPDATE.	*
			 ************************************************/
	    goto Error_Proc ;


	/* create the array */

	if( isc_array_put_slice( isc_status, &DB_info->DB_hdl, &conn->TR_hdl,
	      ( ISC_QUAD * )var->sqldata, &desc, ( void * )src_buf, &src_len ) )
	    goto Error_Proc ;


	return( STATUS_NORMAL ) ;


Error_Proc:
	SQLCODE = isc_sqlcode( isc_status ) ;
	set_dberror( isc_status ) ;

	return( STATUS_ABNORMAL ) ;
}
