/*
 *                       ibconn.c
 *
 *   - The module access to InterBase for the Python.
 *     ( The functions for Connection object )
 *
 *                               Copyright (C) 1997,1998 by RIOS Corporation
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * RIOS Corporation makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.  By use of this software the user agrees to
 * indemnify and hold harmless RIOS Corporation from any  claims or
 * liability for loss arising out of such use.
 */


/* Deallocate the Connection object */

#ifdef __STDC__
static void	*ibconn_dealloc
(
    ibconnobject	*self
)
#else
static void	*ibconn_dealloc( self )
    ibconnobject	*self ;
#endif
{
	/* Disconnect database */

	if( self->open_f )
	    clear_all_database( self ) ;


	/* Free the allocated memory */

	FREE( self->teb_vec ) ;
	FREE( self->i_sqlda ) ;
	FREE( self->o_sqlda ) ;
	FREE( self->results ) ;
	FREE( self->paramts ) ;


	PyMem_DEL( self ) ;

	return ;
}



/* Disconnect from the database */

#ifdef __STDC__
static PyObject	*ibconn_disconnect
(
    ibconnobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibconn_disconnect( self, args )
    ibconnobject	*self ;
    PyObject		*args ;
#endif
{
	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( self, NULL ) ;


	/* Check the arguments */

	if( !PyArg_ParseTuple( args, "" ) )
	    return( NULL ) ;


	/* Disconnect database */

	clear_all_database( self ) ;
	self->open_f = IBOBJ_INIT ;


	Py_INCREF( Py_None ) ;
	return( Py_None ) ;
}



/* Get the attribute of Connection object */

#ifdef __STDC__
static PyObject	*ibconn_getattr
(
    ibconnobject	*self,
    char		*attr_name
)
#else
static PyObject	*ibconn_getattr( self, attr_name )
    ibconnobject	*self ;
    char		*attr_name ;
#endif
{
	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	return( Py_FindMethod( ibconn_methods,
			( PyObject * )self, attr_name ) ) ;
}



/* Create a Cursor object */

#ifdef __STDC__
static PyObject	*ibconn_cursor
(
    ibconnobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibconn_cursor( self, args )
    ibconnobject	*self ;
    PyObject		*args ;
#endif
{
	ibcursobject	*newobj ;
	ibcursobject	*curr_p ;
	char		*alias = NULL ;
	long		 DB_no ;
	DBinformation	*DB_info ;


	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( self, NULL ) ;


	/* Check the arguments */

	if( !PyArg_ParseTuple( args, "|s", &alias ) )
	    return( NULL ) ;


	/* Check the database */

	DB_no = get_database_number( self->DB_num, self->DB_info, alias ) ;
	if( DB_no == STATUS_ABNORMAL )
	{
	    set_pyerror( 0, ERR_NOALIASDB ) ;
	    return( NULL ) ;
	}


	/* Create the cursor object */

	newobj = PyObject_NEW( ibcursobject, &IBcurstype ) ;
	if( newobj == NULL )
	    return( NULL ) ;

	newobj->conn_p = self ;
	sprintf( tempbuff_S, CURSOR_NAMEFMT, curs_num_S ) ;
	newobj->curs_name = strdup( tempbuff_S ) ;
	if( newobj->curs_name == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    Py_XDECREF( newobj ) ;
	    return( NULL ) ;
	}
	curs_num_S ++ ;
	newobj->DB_no = DB_no ;
	newobj->STMT_hdl = NULL ;
	newobj->open_f = IBOBJ_INIT ;
	newobj->arraysize = 1 ;
	newobj->describe = NULL ;
	newobj->prev_p = NULL ;
	newobj->next_p = NULL ;


	/* Allocate the statement handle */

	DB_info = &self->DB_info[DB_no] ;
	isc_dsql_alloc_statement2( isc_status,
			&DB_info->DB_hdl, &newobj->STMT_hdl ) ;
	CHECK_IBERR2( isc_status, NULL ) ;


	/* Set the pointer */

	if( self->first_p )
	{
	    curr_p = self->first_p ;
	    while( curr_p->next_p )
		curr_p = curr_p->next_p ;
	    curr_p->next_p = newobj ;
	    newobj->prev_p = curr_p ;
	}
	else
	    self->first_p = newobj ;


	return( ( PyObject * )newobj ) ;
}



/* Create a Blob object */

#ifdef __STDC__
static PyObject	*ibconn_blob
(
    ibconnobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibconn_blob( self, args )
    ibconnobject	*self ;
    PyObject		*args ;
#endif
{
	ibblobobject	*newobj ;
	char		*table ;
	char		*field ;
	PyObject	*blob = NULL ;


	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( self, NULL ) ;


	/* Get the arguments */

	if( !PyArg_ParseTuple( args, "ss|O", &table, &field, &blob ) )
	    return( NULL ) ;


	/* Create the blob object */

	newobj = new_blob_from_string( self, 0, table, field, blob ) ;
	if( newobj == NULL )
	    return( NULL ) ;


	return( ( PyObject * )newobj ) ;
}



/* Create a Blob object corresponding to the specified database */

#ifdef __STDC__
static PyObject	*ibconn_mblob
(
    ibconnobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibconn_mblob( self, args )
    ibconnobject	*self ;
    PyObject		*args ;
#endif
{
	ibblobobject	*newobj ;
	char		*alias ;
	char		*table ;
	char		*field ;
	PyObject	*blob = NULL ;
	long		 DB_no ;


	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( self, NULL ) ;


	/* Get the arguments */

	if( !PyArg_ParseTuple( args, "sss|O", &alias, &table, &field, &blob ) )
	    return( NULL ) ;


	/* Check the database */

	DB_no = get_database_number( self->DB_num, self->DB_info, alias ) ;
	if( DB_no == STATUS_ABNORMAL )
	{
	    set_pyerror( 0, ERR_NOALIASDB ) ;
	    return( NULL ) ;
	}


	/* Create the blob object */

	newobj = new_blob_from_string( self, DB_no, table, field, blob ) ;
	if( newobj == NULL )
	    return( NULL ) ;


	return( ( PyObject * )newobj ) ;
}



/* Commit the current transaction */

#ifdef __STDC__
static PyObject	*ibconn_commit
(
    ibconnobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibconn_commit( self, args )
    ibconnobject	*self ;
    PyObject		*args ;
#endif
{
	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( self, NULL ) ;


	/* Check the arguments */

	if( !PyArg_ParseTuple( args, "" ) )
	    return( NULL ) ;


	/* Close all cursors */

	close_allcursors( self->first_p ) ;


	/* Commit the transaction */

	if( self->TR_hdl )
	{
	    isc_commit_transaction( isc_status, &self->TR_hdl ) ;
	    CHECK_IBERR2( isc_status, NULL ) ;
	}

	Py_INCREF( Py_None ) ;
	return( Py_None ) ;
}



/* Rollback the current transaction */

#ifdef __STDC__
static PyObject	*ibconn_rollback
(
    ibconnobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibconn_rollback( self, args )
    ibconnobject	*self ;
    PyObject		*args ;
#endif
{
	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( self, NULL ) ;


	/* Check the arguments */

	if( !PyArg_ParseTuple( args, "" ) )
	    return( NULL ) ;


	/* Close all cursors */

	close_allcursors( self->first_p ) ;


	/* Rollback the transaction */

	if( self->TR_hdl )
	{
	    isc_rollback_transaction( isc_status, &self->TR_hdl ) ;
	    CHECK_IBERR2( isc_status, NULL ) ;
	}

	Py_INCREF( Py_None ) ;
	return( Py_None ) ;
}



/* Set the transaction options */

#ifdef __STDC__
static PyObject	*ibconn_transoption
(
    ibconnobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibconn_transoption( self, args )
    ibconnobject	*self ;
    PyObject		*args ;
#endif
{
	char		*opts ;
	char		*alias = NULL ;
	long		 DB_no ;
	char		*ptr ;
	char		*buff ;
	char		*tpb ;
	long		 tpb_len ;
	DBinformation	*DB_info ;


	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( self, NULL ) ;


	/* Check the arguments */

	if( !PyArg_ParseTuple( args, "s|s", &opts, &alias ) )
	    return( NULL ) ;


	/* Check the database */

	DB_no = get_database_number( self->DB_num, self->DB_info, alias ) ;
	if( DB_no == STATUS_ABNORMAL )
	{
	    set_pyerror( 0, ERR_NOALIASDB ) ;
	    return( NULL ) ;
	}


	/* Set the transaction options */

	DB_info = &self->DB_info[DB_no] ;

	if( opts && ( int )strlen( opts ) > 0 )
	{
	    /* Copy the option strings to temporary buffer */

	    if( strlen( opts ) > sizeof( tempbuff_S ) - 1 )
	    {
		buff = strdup( opts ) ;
		if( buff == NULL )
		{
		    set_pyerror( 0, ERR_ALLOCMEM ) ;
		    return( NULL ) ;
		}
	    }
	    else
	    {
		buff = tempbuff_S ;
		strcpy( buff, opts ) ;
	    }

	    /* To upper case */

	    for( ptr = buff ; *ptr ; ptr ++ )
		*ptr = UPPER( *ptr ) ;

	    /* Parse the transaction option */

	    if( parse_transopts( buff, &tpb, &tpb_len ) )
	    {
		if( buff != tempbuff_S )	free( buff ) ;
		return( NULL ) ;
	    }

	    if( buff != tempbuff_S )	free( buff ) ;
	    FREE( DB_info->tpb_ptr ) ;
	    DB_info->tpb_ptr = tpb ;
	    DB_info->tpb_len = tpb_len ;
	}
	else
	{
	    FREE( DB_info->tpb_ptr ) ;
	    DB_info->tpb_ptr = NULL ;
	    DB_info->tpb_len = 0 ;
	}


	Py_INCREF( Py_None ) ;
	return( Py_None ) ;
}



/* Start the transaction */

#ifdef __STDC__
static PyObject	*ibconn_transaction
(
    ibconnobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibconn_transaction( self, args )
    ibconnobject	*self ;
    PyObject		*args ;
#endif
{
	char		*opts ;
	char		*alias = NULL ;
	long		 DB_no ;
	char		*ptr ;
	char		*tpb ;
	long		 tpb_len ;
	DBinformation	*DB_info ;


	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( self, NULL ) ;


	/* Check the arguments */

	if( !PyArg_ParseTuple( args, "" ) )
	    return( NULL ) ;


	/* Check the current transaction */

	if( self->TR_hdl )
	{
	    set_pyerror( 0, ERR_TRSTARTED ) ;
	    return( NULL ) ;
	}


	/* Start transaction */

	if( start_transaction( self ) )
	    return( NULL ) ;


	Py_INCREF( Py_None ) ;
	return( Py_None ) ;
}



/* IB cursor type object definition */

static PyTypeObject	IBcurstype =
{
	PyObject_HEAD_INIT( &PyType_Type )
	0,				/* ob_size */
	"ibcursor",			/* tp_name */
	sizeof(ibcursobject),		/* tp_basicsize */
	0,				/* tp_itemsize */
	( destructor )ibcurs_dealloc,	/* tp_dealloc */
	( printfunc )0,			/* tp_print */
	( getattrfunc )ibcurs_getattr,	/* tp_getattr */
	( setattrfunc )ibcurs_setattr,	/* tp_setattr */
	( cmpfunc )0,			/* tp_compare */
	( reprfunc )0,			/* tp_repr */
	0,				/* tp_as_number */
	0,				/* tp_as_sequence */
	0,				/* tp_as_mapping */
	( hashfunc )0,			/* tp_hash */
	( ternaryfunc )0,		/* tp_call */
	( reprfunc )0,			/* tp_str */
	( getattrofunc )0,		/* tp_getattro */
	( setattrofunc )0,		/* tp_setattro */
	0,				/* tp_xxx3 */
	0,				/* tp_xxx4 */
	0				/* tp_doc */
} ;



/* IB blob object definition */

static PyTypeObject	IBblobtype =
{
	PyObject_HEAD_INIT( &PyType_Type )
	0,				/* ob_size */
	"ibblobtype",			/* tp_name */
	sizeof(ibblobobject),		/* tp_basicsize */
	0,				/* tp_itemsize */
	( destructor )ibblob_dealloc,	/* tp_dealloc */
	( printfunc )ibblob_print,	/* tp_print */
	( getattrfunc )ibblob_getattr,	/* tp_getattr */
	( setattrfunc )0,		/* tp_setattr */
	( cmpfunc )0,			/* tp_compare */
	( reprfunc )0,			/* tp_repr */
	0,				/* tp_as_number */
	0,				/* tp_as_sequence */
	0,				/* tp_as_mapping */
	( hashfunc )0,			/* tp_hash */
	( ternaryfunc )0,		/* tp_call */
	( reprfunc )0,			/* tp_str */
	( getattrofunc )0,		/* tp_getattro */
	( setattrofunc )0,		/* tp_setattro */
	0,				/* tp_xxx3 */
	0,				/* tp_xxx4 */
	0				/* tp_doc */
} ;



/* List of methods defined in connection object */

static struct PyMethodDef ibconn_methods[] =
{
{ "blob",	( PyCFunction )ibconn_blob,		METH_VARARGS,	NULL },
{ "mblob",	( PyCFunction )ibconn_mblob,		METH_VARARGS,	NULL },
{ "close",	( PyCFunction )ibconn_disconnect,	METH_VARARGS,	NULL },
{ "cursor",	( PyCFunction )ibconn_cursor,		METH_VARARGS,	NULL },
{ "commit",	( PyCFunction )ibconn_commit,		METH_VARARGS,	NULL },
{ "rollback",	( PyCFunction )ibconn_rollback,		METH_VARARGS,	NULL },
{ "transopts",	( PyCFunction )ibconn_transoption,	METH_VARARGS,	NULL },
{ "transaction",( PyCFunction )ibconn_transaction,	METH_VARARGS,	NULL },
{ NULL,		NULL }
};
