/*
 *                       ibcursor.c
 *
 *   - The module access to InterBase for the Python.
 *     ( The functions for Cursor object )
 *
 *                               Copyright (C) 1997,1998 by RIOS Corporation
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * RIOS Corporation makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.  By use of this software the user agrees to
 * indemnify and hold harmless RIOS Corporation from any  claims or
 * liability for loss arising out of such use.
 */


/* Set the string parameter */

#ifdef __STDC__
static int	set_stringparam
(
    long	 type,
    PyObject	*obj,
    char	*ptr
)
#else
static int	set_stringparam( type, obj, ptr )
    long	 type ;
    PyObject	*obj ;
    char	*ptr ;
#endif
{
	switch( type )
	{
	    case PYOTP_STRING :
		strcpy( ptr, PyString_AS_STRING( (PyStringObject *)obj ) ) ;
		break ;
	    case PYOTP_INTEGER :
		sprintf( ptr, "%ld", PyInt_AS_LONG( ( PyIntObject * )obj ) ) ;
		break ;
	    case PYOTP_FLOAT :
		sprintf( ptr, "%lg",
			PyFloat_AS_DOUBLE( ( PyFloatObject * )obj ) ) ;
		break ;
	    default :
		set_pyerror( 0, ERR_PARAMITEM ) ;
		return( STATUS_ABNORMAL ) ;
	}
	return( STATUS_NORMAL ) ;
}



/* Get the long integer value */

#ifdef __STDC__
static int	get_longvalue
(
    long	 type,
    short	 scale,
    PyObject	*obj,
    long	*value
)
#else
static int	get_longvalue( type, scale, obj, value )
    long	 type ;
    short	 scale ;
    PyObject	*obj ;
    long	*value ;
#endif
{
	double	 dval ;
	double	 ratio ;
	long	 lval ;
	int	 scnt ;


	switch( type )
	{
	    case PYOTP_STRING :
		if( sscanf( PyString_AS_STRING( (PyStringObject *)obj ),
							"%lg", &dval ) != 1 )
		{
		    set_pyerror( 0, ERR_STRTOINT ) ;
		    return( STATUS_ABNORMAL ) ;
		}
		break ;
	    case PYOTP_INTEGER :
		dval = PyInt_AS_LONG( ( PyIntObject * )obj ) ;
		break ;
	    case PYOTP_FLOAT :
		dval = PyFloat_AS_DOUBLE( ( PyFloatObject * )obj ) ;
		break ;
	    default :
		set_pyerror( 0, ERR_PARAMITEM ) ;
		return( STATUS_ABNORMAL ) ;
	}

	ratio = 1 ;
	for( scnt = 0 ; scnt > scale ; scnt -- )
	    ratio *= 10 ;
	sprintf( tempbuff_S, "%lf", dval * ratio ) ;
	sscanf( tempbuff_S, "%ld", &lval ) ;

	*value = lval ;

	return( STATUS_NORMAL ) ;
}



/* Get the floating point value */

#ifdef __STDC__
static int	get_realvalue
(
    long	 type,
    PyObject	*obj,
    double	*value
)
#else
static int	get_realvalue( type, obj, value )
    long	 type ;
    PyObject	*obj ;
    double	*value ;
#endif
{
	double	 dval ;


	switch( type )
	{
	    case PYOTP_STRING :
		if( sscanf( PyString_AS_STRING( (PyStringObject *)obj ),
							"%lg", &dval ) != 1 )
		{
		    set_pyerror( 0, ERR_STRTOREAL ) ;
		    return( STATUS_ABNORMAL ) ;
		}
		break ;
	    case PYOTP_INTEGER :
		dval = PyInt_AS_LONG( ( PyIntObject * )obj ) ;
		break ;
	    case PYOTP_FLOAT :
		dval = PyFloat_AS_DOUBLE( ( PyFloatObject * )obj ) ;
		break ;
	    default :
		set_pyerror( 0, ERR_PARAMITEM ) ;
		return( STATUS_ABNORMAL ) ;
	}

	*value = dval ;

	return( STATUS_NORMAL ) ;
}



/* Get the date */

#ifdef __STDC__
static int	get_datevalue
(
    long	 type,
    PyObject	*obj,
    ISC_QUAD	*value
)
#else
static int	get_datevalue( type, obj, value )
    long	 type ;
    PyObject	*obj ;
    ISC_QUAD	*value ;
#endif
{
	switch( type )
	{
	    case PYOTP_IBDATE :
		isc_encode_date( &((( ibdateobject * )obj)->dtime), value ) ;
		break ;
	    default :
		set_pyerror( 0, ERR_PARAMITEM ) ;
		return( STATUS_ABNORMAL ) ;
	}

	return( STATUS_NORMAL ) ;
}



/* Get the blob */

#ifdef __STDC__
static int	get_blobvalue
(
    ibcursobject	*curs,
    long		 type,
    PyObject		*obj,
    int			 text_f,
    ISC_QUAD		*blob_id
)
#else
static int	get_blobvalue( curs, type, obj, text_f, blob_id )
    ibcursobject	*curs ;
    long		 type ;
    PyObject		*obj ;
    int			 text_f ;
    ISC_QUAD		*blob_id ;
#endif
{
	char		*ptr ;
	ibblobobject	*blob ;


	switch( type )
	{
	    case PYOTP_STRING :
		ptr = PyString_AS_STRING( ( PyStringObject * )obj ) ;
		if( create_blob( curs, text_f, NULL,
				PyString_Size( obj ), ptr, blob_id ) )
		    return( STATUS_ABNORMAL ) ;
		break ;
	    case PYOTP_IBBLOB :
		blob = ( ibblobobject * )obj ;
		if( create_blob( curs, text_f, &blob->desc,
				blob->size, blob->blob, blob_id ) )
		    return( STATUS_ABNORMAL ) ;
		break ;
	    default :
		set_pyerror( 0, ERR_PARAMITEM ) ;
		return( STATUS_ABNORMAL ) ;
	}
	return( STATUS_NORMAL ) ;
}



/* Get the array */

#ifdef __STDC__
static int	get_arrayvalue
(
    ibcursobject	*curs,
    long		 type,
    PyObject		*obj,
    XSQLVAR		*var
)
#else
static int	get_arrayvalue( curs, type, obj, var )
    ibcursobject	*curs ;
    long		 type ;
    PyObject		*obj ;
    XSQLVAR		*var ;
#endif
{
	char		*ptr ;

/**********************************
 * This function is not used now! *
 **********************************/

	switch( type )
	{
	    case PYOTP_STRING :
		ptr = PyString_AS_STRING( ( PyStringObject * )obj ) ;
		if( create_array( curs, var, PyString_Size( obj ), ptr ) )
		    return( STATUS_ABNORMAL ) ;
		break ;
	    default :
		set_pyerror( 0, ERR_PARAMITEM ) ;
		return( STATUS_ABNORMAL ) ;
	}
	return( STATUS_NORMAL ) ;
}



/* Set the input parameters */

#ifdef __STDC__
static int	set_inputparams
(
    ibcursobject	*self,
    PyObject		*params,
    long		 items,
    PyObject		*( * func_getitem )()
)
#else
static int	set_inputparams( self, params, items, func_getitem )
    ibcursobject	*self ;
    PyObject		*params ;
    long		 items ;
    PyObject		*( * func_getitem )() ;
#endif
{
	ibconnobject	*conn = self->conn_p ;
	long		 in_params ;
	long		 count ;
	PyObject	*obj ;
	long		 objnum ;
	void		*data ;
	long		 type ;
	XSQLVAR		*var ;
	short		 dtp ;
	long		 length ;
	long		 alignment ;
	long		 offset ;
	long		 lvalue ;
	double		 dvalue ;
	double		 dcheck ;
	VARY		*vary ;



	/* Check the number of parameters */

	in_params = conn->i_sqlda->sqld ;
	if( in_params != items )
	{
	    sprintf( tempbuff_S, ERR_UNMATCHNO, in_params ) ;
	    set_pyerror( 0, tempbuff_S ) ;
	    return( STATUS_ABNORMAL ) ;
	}


	/* Get the parameters */

	for( count = 0, offset = 0 ; count < items ; count ++ )
	{
	    obj = func_getitem( params, count ) ;
	    if( obj == Py_None )
		type = PYOTP_NONEOBJ ;
	    else if( PyString_Check( obj ) )
		type = PYOTP_STRING ;
	    else if( PyInt_Check( obj ) )
		type = PYOTP_INTEGER ;
	    else if( PyFloat_Check( obj ) )
		type = PYOTP_FLOAT ;
	    else if( IS_DATEOBJECT( obj ) )
		type = PYOTP_IBDATE ;
	    else if( IS_BLOBOBJECT( obj ) )
		type = PYOTP_IBBLOB ;
	    else
		type = PYOTP_OTHERS ;


	    /* Convert the data type for InterBase */

	    var = &conn->i_sqlda->sqlvar[count] ;

	    if( type != PYOTP_NONEOBJ )
	    {
		dtp = var->sqltype & ~1 ;		/* Erase null flag */
		alignment = var->sqllen ;

		switch( dtp )
		{
		    case SQL_TEXT :
			alignment = 1 ;
			offset = ALIGN( offset, alignment ) ;
			var->sqldata = ( char * )( conn->paramts + offset ) ;
			if( set_stringparam( type, obj, var->sqldata ) )
			    return( STATUS_ABNORMAL ) ;
			var->sqllen = strlen( var->sqldata ) ;
			offset += var->sqllen + 1 ;
			break ;

		    case SQL_VARYING :
			alignment = sizeof( short ) ;
			offset = ALIGN( offset, alignment ) ;
			var->sqldata = ( char * )( conn->paramts + offset ) ;
			vary = ( VARY * )var->sqldata ;
			if( set_stringparam( type, obj, vary->vary_string ) )
			    return( STATUS_ABNORMAL ) ;
			vary->vary_length = strlen( vary->vary_string ) ;
			offset += vary->vary_length + sizeof( short ) ;
			break ;

		    case SQL_SHORT :
			offset = ALIGN( offset, alignment ) ;
			var->sqldata = ( char * )( conn->paramts + offset ) ;
			if( get_longvalue( type, var->sqlscale, obj, &lvalue ) )
			    return( STATUS_ABNORMAL ) ;
			if( lvalue < SHRT_MIN || lvalue > SHRT_MAX )
			{
			    set_pyerror( 0, ERR_SHORTOVER ) ;
			    return( STATUS_ABNORMAL ) ;
			}
			*( short * )var->sqldata = lvalue ;
			offset += alignment ;
			break ;

		    case SQL_LONG :
			offset = ALIGN( offset, alignment ) ;
			var->sqldata = ( char * )( conn->paramts + offset ) ;
			if( get_longvalue( type, var->sqlscale, obj, &lvalue ) )
			    return( STATUS_ABNORMAL ) ;
			*( long * )var->sqldata = lvalue ;
			offset += alignment ;
			break ;

		    case SQL_FLOAT :
			offset = ALIGN( offset, alignment ) ;
			var->sqldata = ( char * )( conn->paramts + offset ) ;
			if( get_realvalue( type, obj, &dvalue ) )
			    return( STATUS_ABNORMAL ) ;
			if( dvalue >= 0.0 )
			    dcheck = dvalue ;
			else
			    dcheck = dvalue * -1 ;

			if( dcheck != 0.0
			&& ( dcheck < FLT_MIN || dcheck > FLT_MAX ) )
			{
			    set_pyerror( 0, ERR_FLOATOVER ) ;
			    return( STATUS_ABNORMAL ) ;
			}
			*( float * )var->sqldata = dvalue ;
			offset += alignment ;
			break ;

		    case SQL_DOUBLE :
			offset = ALIGN( offset, alignment ) ;
			var->sqldata = ( char * )( conn->paramts + offset ) ;
			if( get_realvalue( type, obj, &dvalue ) )
			    return( STATUS_ABNORMAL ) ;
			*( double * )var->sqldata = dvalue ;
			offset += alignment ;
			break ;

		    case SQL_DATE :
			offset = ALIGN( offset, alignment ) ;
			var->sqldata = ( char * )( conn->paramts + offset ) ;
			if( get_datevalue( type, obj,
						( ISC_QUAD * )var->sqldata ) )
			    return( STATUS_ABNORMAL ) ;
			offset += alignment ;
			break ;

		    case SQL_BLOB :
			offset = ALIGN( offset, alignment ) ;
			var->sqldata = ( char * )( conn->paramts + offset ) ;
			if( get_blobvalue( self, type, obj,
				( var->sqlsubtype == BLOB_SBTP_TEXT ? 1 : 0 ),
				( ISC_QUAD * )var->sqldata ) )
			    return( STATUS_ABNORMAL ) ;
			offset += alignment ;
			break ;

		    case SQL_ARRAY :
/* Not support now
			offset = ALIGN( offset, alignment ) ;
			var->sqldata = ( char * )( conn->paramts + offset ) ;
			if( get_arrayvalue( self, type, obj, var ) )
			    return( STATUS_ABNORMAL ) ;
			offset += alignment ;
			break ;
*/
			set_pyerror( 0, ERR_NOSPARRAY ) ;
			return( STATUS_ABNORMAL ) ;

		    default :
			set_pyerror( 0, ERR_UNSPPORTD ) ;
			return( STATUS_ABNORMAL ) ;
		}

		if( var->sqltype & 1 )
		{
		    offset = ALIGN( offset, sizeof( short ) ) ;
		    var->sqlind = ( short * )( conn->paramts + offset ) ;
		    *var->sqlind = 0 ;
		    offset += sizeof( short ) ;
		}
	    }
	    else
	    {
		if( var->sqltype & 1 )
		{
		    var->sqldata = NULL ;
		    offset = ALIGN( offset, sizeof( short ) ) ;
		    var->sqlind = ( short * )( conn->paramts + offset ) ;
		    *var->sqlind = -1 ;
		    offset += sizeof( short ) ;
		}
		else
		{
		    set_pyerror( 0, ERR_SETNONULL ) ;
		    return( STATUS_ABNORMAL ) ;
		}
	    }
	}

	return( STATUS_NORMAL ) ;
}



/* Set up the output XSQLDA */

#ifdef __STDC__
static void	set_outputsqlda
(
    ibcursobject	*self,
    XSQLDA		*sqlda
)
#else
static void	set_outputsqlda( self, sqlda )
    ibcursobject	*self ;
    XSQLDA		*sqlda ;
#endif
{
	ibconnobject	*conn = self->conn_p ;
	long		 cols ;
	long		 count ;
	XSQLVAR		*var ;
	short		 dtp ;
	long		 length ;
	long		 alignment ;
	long		 offset ;



	cols = sqlda->sqld ;
	for( var = sqlda->sqlvar, offset = 0, count = 0 ;
				count < cols ; var++, count ++ )
	{
	    length = alignment = var->sqllen ;
	    dtp = var->sqltype & ~1 ;

	    if( dtp == SQL_TEXT )
		alignment = 1 ;
	    else if( dtp == SQL_VARYING )
	    {
		length += sizeof( short ) ;
		alignment = sizeof( short ) ;
	    }
	    offset = ALIGN( offset, alignment ) ;
	    var->sqldata = ( char * )( conn->results + offset ) ;
	    offset += length ;
	    offset = ALIGN( offset, sizeof( short ) ) ;
	    var->sqlind = ( short * )( conn->results + offset ) ;
	    offset += sizeof( short ) ;
	}

	return ;
}



/* Set the description */

#ifdef __STDC__
static int	set_description
(
    XSQLDA		 *sqlda,
    PyObject		**desc
)
#else
static int	set_description( sqlda, desc )
    XSQLDA		 *sqlda ;
    PyObject		**desc ;
#endif
{
	long		 cols ;
	long		 count ;
	XSQLVAR		*var ;
	short		 dtp ;
	PyObject	*tuple ;
	PyObject	*coltuple ;
	long		 objcnt ;



	/* Free the previous description */

	if( *desc )
	{
	    free_description( *desc ) ;
	    *desc = NULL ;
	}


	/* Check the number of result columns */

	cols = sqlda->sqld ;
	if( !cols )
	{
	    *desc = NULL ;
	    return( STATUS_NORMAL ) ;
	}


	/* Create tuple object */

	if( ( tuple = PyTuple_New( cols ) ) == NULL )
	    return( STATUS_ABNORMAL ) ;


	/* Check the number of result columns */

	for( count = 0 ; count < cols ; count ++ )
	{
	    if( ( coltuple = PyTuple_New( COLATTR_NUMBER ) ) == NULL )
		return( STATUS_ABNORMAL ) ;
	    objcnt = 0 ;

	    var = &sqlda->sqlvar[count] ;
	    dtp = var->sqltype & ~1 ;

	    PyTuple_SetItem( coltuple, objcnt ++,	/* columns name */
			PyString_FromStringAndSize( var->sqlname,
						var->sqlname_length ) ) ;

	    PyTuple_SetItem( coltuple, objcnt ++,	/* type code */
			PyInt_FromLong( ( long )( var->sqltype & ~1 ) ) ) ;

	    PyTuple_SetItem( coltuple, objcnt ++,	/* display size */
			PyInt_FromLong( ( long )var->sqllen ) ) ;

	    if( dtp == SQL_VARYING )
		PyTuple_SetItem( coltuple, objcnt ++,	/* internal size */
			PyInt_FromLong( (long)var->sqllen + sizeof(short) ) ) ;
	    else
		PyTuple_SetItem( coltuple, objcnt ++,	/* internal size */
			PyInt_FromLong( ( long )var->sqllen ) ) ;

	    PyTuple_SetItem( coltuple, objcnt ++,	/* precision */
			PyInt_FromLong( 0L ) ) ;

	    PyTuple_SetItem( coltuple, objcnt ++,	/* scale */
			PyInt_FromLong( ( long )var->sqlscale ) ) ;

	    PyTuple_SetItem( coltuple, objcnt,		/* null OK */
			PyInt_FromLong( ( long )( var->sqltype & 1 ) ) ) ;

	    PyTuple_SetItem( tuple, count, coltuple ) ;
	}

	*desc = tuple ;

	return( STATUS_NORMAL ) ;
}



/* Execute SQL with parameters */

#ifdef __STDC__
static int	execute_withparams
(
    ibcursobject	*self,
    PyObject		*params,
    long		 items,
    PyObject		*( * func_getitem )()
)
#else
static int	execute_withparams( self, params, items, func_getitem )
    ibcursobject	*self ;
    PyObject		*params ;
    long		 items ;
    PyObject		*( * func_getitem )() ;
#endif
{
	ibconnobject	*conn = self->conn_p ;
	long		 count ;
	PyObject	*obj ;
	long		 objnum ;


	/* Check the number of parameters */

	if( items == 0 )
	    return( STATUS_NORMAL ) ;


	/* Check the first object type of the parameters */

	obj = func_getitem( params, 0 ) ;
	if( PyTuple_Check( obj ) || PyList_Check( obj ) )
	{
	    for( count = 0 ; count < items ; count ++ )
	    {
		/* Set the input parameters */

		obj = func_getitem( params, count ) ;
		if( PyTuple_Check( obj ) )
		{
		    objnum = PyTuple_Size( obj ) ;
		    if( set_inputparams( self, obj, objnum, PyTuple_GetItem ) )
			return( STATUS_ABNORMAL ) ;
		}
		else if( PyList_Check( obj ) )
		{
		    objnum = PyList_Size( obj ) ;
		    if( set_inputparams( self, obj, objnum, PyList_GetItem ) )
			return( STATUS_ABNORMAL ) ;
		}
		else
		{
		    set_pyerror( 0, ERR_PARAMTYPE ) ;
		    return( STATUS_ABNORMAL ) ;
		}


		/* Execute SQL statement */

		isc_dsql_execute2( isc_status, &conn->TR_hdl,
				&self->STMT_hdl, 1, conn->i_sqlda, NULL ) ;
		CHECK_IBERR1( isc_status ) ;
	    }
	}
	else
	{
	    /* Set the input parameters */

	    if( set_inputparams( self, params, items, func_getitem ) )
		return( STATUS_ABNORMAL ) ;


	    /* Execute SQL statement */

	    isc_dsql_execute2( isc_status, &conn->TR_hdl,
				&self->STMT_hdl, 1, conn->i_sqlda, NULL ) ;
	    CHECK_IBERR1( isc_status ) ;
	}

	return( STATUS_NORMAL ) ;


Error_Proc:
	set_dberror( isc_status ) ;
	return( STATUS_ABNORMAL ) ;
}



/* Check the input parameters */

#ifdef __STDC__
static int	check_inparams
(
    int			 exec_f,
    ibcursobject	*self,
    PyObject		*params
)
#else
static int	check_inparams( exec_f, self, params )
    int			 exec_f ;
    ibcursobject	*self ;
    PyObject		*params ;
#endif
{
	long		 items ;
	PyObject	*( * func_getitem )() ;



	/* Check the parameter */

	if( params == NULL )
	{
	    set_pyerror( 0, ERR_NOINPARAM ) ;
	    return( STATUS_ABNORMAL ) ;
	}

	if( PyTuple_Check( params ) )
	{
	    items = PyTuple_Size( params ) ;
	    func_getitem = PyTuple_GetItem ;
	}
	else if( PyList_Check( params ) )
	{
	    items = PyList_Size( params ) ;
	    func_getitem = PyList_GetItem ;
	}
	else
	{
	    set_pyerror( 0, ERR_PARAMTYPE ) ;
	    return( STATUS_ABNORMAL ) ;
	}


	/* Execute specified process */

	switch( exec_f )
	{
	    case EXECF_EXECDML :
		return( execute_withparams( self, params,
					items, func_getitem ) ) ;
	    case EXECF_SETPARM :
		return( set_inputparams( self, params,
					items, func_getitem ) ) ;
	    default :
		return( STATUS_ABNORMAL ) ;
	}
}



/* Fetch one record from the results */

#ifdef __STDC__
static PyObject	*fetch_onerow
(
    ibcursobject	*self,
    long		*status
)
#else
static PyObject	*fetch_onerow( self, status )
    ibcursobject	*self ;
    long		*status ;
#endif
{
	ibconnobject	*conn = self->conn_p ;
	long		 cols ;
	long		 count ;
	long		 scnt ;
	XSQLVAR		*var ;
	PyObject	*tuple ;
	long		 dtp ;
	VARY		*vary ;
	double		 ratio ;
	double		 dval ;
	struct tm	 tms ;
	ibdateobject	*newobj ;
	ibblobobject	*newblob ;
	long		 length ;
	char		*blobd ;
	ISC_BLOB_DESC	 desc ;



	/* Initialize */

	*status = STATUS_ABNORMAL ;


	/* Fetch one row */

	SQLCODE = isc_dsql_fetch( isc_status,
				&self->STMT_hdl, 1, conn->o_sqlda ) ;
	if( SQLCODE == SQLCODE_NOMORE )
	{
	    Py_INCREF( Py_None ) ;
	    *status = SQLCODE_NOMORE ;
	    return( Py_None ) ;
	}
	CHECK_IBERR1( isc_status ) ;


	/* Create the result tuple object */

	cols = conn->o_sqlda->sqld ;
	if( ( tuple = PyTuple_New( cols ) ) == NULL )
	    return( NULL ) ;


	/* Create the result objects for each columns */

	for( count = 0 ; count < cols ; count ++ )
	{
	    var = &conn->o_sqlda->sqlvar[count] ;
	    dtp = var->sqltype & ~1 ;

	    /* Check if column is null */

	    if( ( var->sqltype & 1 ) && ( *var->sqlind < 0 ) )
	    {
		Py_INCREF( Py_None ) ;
		PyTuple_SetItem( tuple, count, Py_None ) ;
	    }
	    else
	    {
		/* Set the column value to the result tuple */

		switch( dtp )
		{
		    case SQL_TEXT :
			PyTuple_SetItem( tuple, count,
				PyString_FromStringAndSize(
					var->sqldata, var->sqllen ) ) ;
			break ;

		    case SQL_VARYING :
			vary = ( VARY * )var->sqldata ;
			PyTuple_SetItem( tuple, count,
				PyString_FromStringAndSize( vary->vary_string,
						vary->vary_length ) ) ;
			break ;

		    case SQL_SHORT :
			if( var->sqlscale < 0 )
			{
			    ratio = 1 ;
			    for( scnt = 0 ; scnt > var->sqlscale ; scnt -- )
				ratio *= 10 ;
			    dval = ( double )*( short * )var->sqldata / ratio ;
			    PyTuple_SetItem( tuple, count,
					PyFloat_FromDouble( dval ) ) ;
			}
			else
			{
			    PyTuple_SetItem( tuple, count, PyInt_FromLong(
					( long )*( short * )var->sqldata ) ) ;
			}
			break ;

		    case SQL_LONG :
			if( var->sqlscale < 0 )
			{
			    ratio = 1 ;
			    for( scnt = 0 ; scnt > var->sqlscale ; scnt -- )
				ratio *= 10 ;
			    dval = ( double )*( long * )var->sqldata / ratio ;
			    PyTuple_SetItem( tuple, count,
					PyFloat_FromDouble( dval ) ) ;
			}
			else
			{
			    PyTuple_SetItem( tuple, count,
				PyInt_FromLong( *( long * )var->sqldata ) ) ;
			}
			break ;

		    case SQL_FLOAT :
			PyTuple_SetItem( tuple, count,
				PyFloat_FromDouble( ( double )*( float * )
							var->sqldata ) ) ;
			break ;

		    case SQL_DOUBLE :
			PyTuple_SetItem( tuple, count,
				PyFloat_FromDouble( *( double * )
							var->sqldata ) ) ;
			break ;

		    case SQL_DATE :
			isc_decode_date( ( ISC_QUAD * )var->sqldata, &tms ) ;
			newobj = new_date_object( &tms ) ;
			if( newobj == NULL )
			    return( NULL ) ;
			PyTuple_SetItem( tuple, count, ( PyObject * )newobj ) ;
			break ;

		    case SQL_BLOB :
			if( retrieve_blob( self, var, &length, &blobd, &desc ) )
			    return( NULL ) ;
			newblob = new_blob_object( length, blobd, &desc ) ;
			if( newblob == NULL )
			    return( NULL ) ;
			PyTuple_SetItem( tuple, count, ( PyObject * )newblob ) ;
			break ;

		    case SQL_ARRAY :
/* Not support now
			if( retrieve_array( self, var, &length, &blobd ) )
			    return( NULL ) ;
			PyTuple_SetItem( tuple, count,
				PyString_FromStringAndSize( blobd, length ) ) ;
			free( blobd ) ;
			break ;
*/
			set_pyerror( 0, ERR_NOSPARRAY ) ;
			return( NULL ) ;

		    default :
			set_pyerror( 0, ERR_UNSPPORTD ) ;
			return( NULL ) ;
		}
	    }
	}

	*status = STATUS_NORMAL ;
	return( tuple ) ;


Error_Proc:
	set_dberror( isc_status ) ;
	return( NULL ) ;
}



/* Deallocate the Cursor object */

#ifdef __STDC__
static void	*ibcurs_dealloc
(
    ibcursobject	*self
)
#else
static void	*ibcurs_dealloc( self )
    ibcursobject	*self ;
#endif
{
	ibcursobject	*prev_p ;
	ibcursobject	*next_p ;


	/* Close the cursor */

	if( self->open_f )
	    isc_dsql_free_statement( isc_status, &self->STMT_hdl, DSQL_close ) ;

	isc_dsql_free_statement( isc_status, &self->STMT_hdl, DSQL_drop ) ;
 
	free( self->curs_name ) ;
	if( self->describe )
	    free_description( self->describe ) ;

	prev_p = self->prev_p ;
	next_p = self->next_p ;

	if( prev_p )
	{
	    prev_p->next_p = next_p ;
	}
	else
	{
	    if( self->conn_p )
		self->conn_p->first_p = next_p ;
	}

	if( next_p )
	    next_p->prev_p = prev_p ;

	PyMem_DEL( self ) ;

	return ;
}



/* Get the attribute of the Cursor object */

#ifdef __STDC__
static PyObject	*ibcurs_getattr
(
    ibcursobject	*self,
    char		*attr_name
)
#else
static PyObject	*ibcurs_getattr( self, attr_name )
    ibcursobject	*self ;
    char		*attr_name ;
#endif
{
	ibconnobject	*conn = self->conn_p ;


	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( conn, NULL ) ;


	if( !strcmp( attr_name, CURATTR_ARRAY ) )
	{
	    return( PyInt_FromLong( self->arraysize ) ) ;
	}
	else if( !strcmp( attr_name, CURATTR_DESCS ) )
	{
	    if( self->describe )
	    {
		Py_INCREF( self->describe ) ;
		return( self->describe ) ;
	    }
	    else
	    {
		Py_INCREF( Py_None ) ;
		return( Py_None ) ;
	    }
	}

	return( Py_FindMethod( ibcurs_methods,
			( PyObject * )self, attr_name ) ) ;
}



/* Set the attribute of the Cursor object */

#ifdef __STDC__
static int	ibcurs_setattr
(
    ibcursobject	*self,
    char		*attr_name,
    PyObject		*args
)
#else
static int	ibcurs_setattr( self, attr_name, args )
    ibcursobject	*self ;
    char		*attr_name ;
    PyObject		*args ;
#endif
{
	long	 	 larg ;
	ibconnobject	*conn = self->conn_p ;


	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( conn, STATUS_ABNORMAL ) ;


	if( !strcmp( attr_name, CURATTR_ARRAY ) )
	{
	    /* Get the arguments */

	    if( PyInt_Check( args ) )
		larg = PyInt_AS_LONG( ( PyIntObject * )args ) ;
	    else
	    {
		set_pyerror( 0, ERR_ARRAYSINT ) ;
		return( STATUS_ABNORMAL ) ;
	    }

	    if( larg <= 0 )
	    {
		set_pyerror( 0, ERR_ARRAYSIZE ) ;
		return( STATUS_ABNORMAL ) ;
	    }

	    self->arraysize = larg ;
	    return( STATUS_NORMAL ) ;
	}

	return( STATUS_ABNORMAL ) ;
}



/* Close the cursor */

#ifdef __STDC__
static PyObject	*ibcurs_close
(
    ibcursobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibcurs_close( self, args )
    ibcursobject	*self ;
    PyObject		*args ;
#endif
{
	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( self->conn_p, NULL ) ;


	/* Close the cursors */

	if( self->open_f )
	{
	    isc_dsql_free_statement( isc_status,
					&self->STMT_hdl, DSQL_close ) ;
	    self->open_f = IBOBJ_INIT ;
	}

	if( self->describe )
	{
	    Py_DECREF( self->describe ) ;
	    self->describe = NULL ;
	}


	Py_INCREF( Py_None ) ;
	return( Py_None ) ;
}



/* Execute the SQL statement */

#ifdef __STDC__
static PyObject	*ibcurs_execute
(
    ibcursobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibcurs_execute( self, args )
    ibcursobject	*self ;
    PyObject		*args ;
#endif
{
	PyObject	*params = NULL ;
	char		*sql ;
	ibconnobject	*conn = self->conn_p ;
	long		 in_params ;
	long		 cols ;
	long		 items ;
	long		 length ;
	long		 statement ;
	PyObject	*ret_obj ;


	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( conn, NULL ) ;


	/* Get the arguments */

	if( !PyArg_ParseTuple( args, "s|O", &sql, &params ) )
	    return( NULL ) ;


	/* Close the cursors */

	if( self->open_f )
	{
	    isc_dsql_free_statement( isc_status,
					&self->STMT_hdl, DSQL_close ) ;
	    self->open_f = IBOBJ_INIT ;
	}


	/* Start the transaction */

	if( !conn->TR_hdl )
	{
	    if( start_transaction( conn ) )
		return( NULL ) ;
	}


	/* Prepare query */

	isc_dsql_prepare( isc_status, &conn->TR_hdl,
				&self->STMT_hdl, 0, sql, 1, conn->o_sqlda ) ;
	CHECK_IBERR1( isc_status ) ;


	/* Get the statement type */

	isc_dsql_sql_info( isc_status, &self->STMT_hdl,
		sizeof( isc_info_stmt ), isc_info_stmt,
		sizeof( isc_info_buff ), isc_info_buff	) ;
	CHECK_IBERR1( isc_status ) ;

	if( isc_info_buff[0] == isc_info_sql_stmt_type )
	{
	    length = isc_vax_integer( &isc_info_buff[1], 2 ) ;
	    statement = isc_vax_integer( &isc_info_buff[3], ( short )length ) ;
	}
	else
	    statement = 0 ;


	/* Describe the parameters */

	isc_dsql_describe_bind( isc_status,
				&self->STMT_hdl, 1, conn->i_sqlda ) ;
	CHECK_IBERR1( isc_status ) ;

	isc_dsql_describe( isc_status, &self->STMT_hdl, 1, conn->o_sqlda ) ;
	CHECK_IBERR1( isc_status ) ;


	/* Get the number of parameters and reallocate the SQLDA */

	in_params = conn->i_sqlda->sqld ;
	if( conn->i_sqlda->sqln < in_params )
	{
	    free( conn->i_sqlda ) ;
	    conn->i_sqlda = alloc_sqlda( in_params ) ;
	    if( conn->i_sqlda == NULL )		return( NULL ) ;

	    /* Describe again */

	    isc_dsql_describe_bind( isc_status,
				&self->STMT_hdl, 1, conn->i_sqlda ) ;
	    CHECK_IBERR1( isc_status ) ;
	}


	/* Get the size of parameters buffer and reallocate it */

	if( in_params )
	{
	    length = calculate_buffsize( conn->i_sqlda ) ;
	    if( length > conn->prmsize )
	    {
		conn->paramts = realloc_buffer( conn->paramts, length ) ;
		if( conn->paramts == NULL )
		    return( NULL ) ;
		conn->prmsize = length ;
	    }
	}


	/* Execute the SQL statement if it is not query */

	if( !conn->o_sqlda->sqld )
	{
	    if( statement == isc_info_sql_stmt_start_trans )
	    {
		set_pyerror( 0, ERR_USETRANS ) ;
		return( NULL ) ;
	    }
	    else if( statement == isc_info_sql_stmt_commit )
	    {
		set_pyerror( 0, ERR_USECOMMIT ) ;
		return( NULL ) ;
	    }
	    else if( statement == isc_info_sql_stmt_rollback )
	    {
		set_pyerror( 0, ERR_USEROLLBK ) ;
		return( NULL ) ;
	    }
	    else
	    {
		if( in_params )
		{
		    if( check_inparams( EXECF_EXECDML, self, params ) )
			return( NULL ) ;
		}
		else
		{
		    isc_dsql_execute2( isc_status, &conn->TR_hdl,
					&self->STMT_hdl, 1, NULL, NULL ) ;
		    CHECK_IBERR1( isc_status ) ;
		}
	    }
	}


	/* Open cursor if the SQL statement is query */

	else
	{
	    /* Get the number of columns and reallocate the SQLDA */

	    cols = conn->o_sqlda->sqld ;
	    if( conn->o_sqlda->sqln < cols )
	    {
		free( conn->o_sqlda ) ;
		conn->o_sqlda = alloc_sqlda( cols ) ;
		if( conn->o_sqlda == NULL )		return( NULL ) ;

		/* Describe again */

		isc_dsql_describe( isc_status,
				&self->STMT_hdl, 1, conn->o_sqlda ) ;
		CHECK_IBERR1( isc_status ) ;
	    }


	    if( in_params )
	    {
		if( check_inparams( EXECF_SETPARM, self, params ) )
		    return( NULL ) ;
	    }


	    /* Declare cursor */

	    isc_dsql_set_cursor_name( isc_status,
				&self->STMT_hdl, self->curs_name, 0 ) ;
	    CHECK_IBERR1( isc_status ) ;


	    /* Open cursor */

	    if( in_params )
	    {
		isc_dsql_execute2( isc_status, &conn->TR_hdl,
				&self->STMT_hdl, 1, conn->i_sqlda, NULL ) ;
		CHECK_IBERR1( isc_status ) ;
	    }
	    else
	    {
		isc_dsql_execute2( isc_status, &conn->TR_hdl,
				&self->STMT_hdl, 1, NULL, NULL ) ;
		CHECK_IBERR1( isc_status ) ;
	    }
	    self->open_f = IBOBJ_OPEN ;


	    /* Get the size of results buffer and reallocate it */

	    length = calculate_buffsize( conn->o_sqlda ) ;
	    if( length > conn->ressize )
	    {
		conn->results = realloc_buffer( conn->results, length ) ;
		if( conn->results == NULL )
		    return( NULL ) ;
		conn->ressize = length ;
	    }
	}


	/* Set the description attributes */

	if( set_description( conn->o_sqlda, &self->describe ) )
	    return( NULL ) ;


	/* Set the return object */

	if( statement == isc_info_sql_stmt_select
	||  statement == isc_info_sql_stmt_select_for_upd )
	{
	    Py_INCREF( Py_None ) ;
	    ret_obj = Py_None ;
	}
	else
	{
	    if( statement == isc_info_sql_stmt_ddl )
		ret_obj = PyInt_FromLong( STATEMENT_DDL ) ;
	    else
		ret_obj = PyInt_FromLong( STATEMENT_DML ) ;
	}

	return( ret_obj ) ;


Error_Proc:
	set_dberror( isc_status ) ;
	return( NULL ) ;
}



/* Fetch the next one row */

#ifdef __STDC__
static PyObject	*ibcurs_fetchone
(
    ibcursobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibcurs_fetchone( self, args )
    ibcursobject	*self ;
    PyObject		*args ;
#endif
{
	ibconnobject	*conn = self->conn_p ;
	long		 status ;



	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( conn, NULL ) ;


	/* Check if open cursor */

	if( !self->open_f )
	{
	    set_pyerror( 0, ERR_NOOPENCUR ) ;
	    return( NULL ) ;
	}


	/* Check the arguments */

	if( !PyArg_ParseTuple( args, "" ) )
	    return( NULL ) ;


	/* Describe output SQLDA */

	isc_dsql_describe( isc_status, &self->STMT_hdl, 1, conn->o_sqlda ) ;
	CHECK_IBERR2( isc_status, NULL ) ;


	/* Set the output SQLDA */

	set_outputsqlda( self, conn->o_sqlda ) ;


	/* Fetch one row and return result */

	return( fetch_onerow( self, &status ) ) ;
}



/* Fetch the specified numbers of next rows */

#ifdef __STDC__
static PyObject	*ibcurs_fetchmany
(
    ibcursobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibcurs_fetchmany( self, args )
    ibcursobject	*self ;
    PyObject		*args ;
#endif
{
	ibconnobject	*conn = self->conn_p ;
	long		 status ;
	long		 rows = 0 ;
	long		 count ;
	PyObject	*tuple ;
	PyObject	*list ;



	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( conn, NULL ) ;


	/* Check if open cursor */

	if( !self->open_f )
	{
	    set_pyerror( 0, ERR_NOOPENCUR ) ;
	    return( NULL ) ;
	}


	/* Check the arguments */

	if( !PyArg_ParseTuple( args, "|l", &rows ) )
	    return( NULL ) ;
	if( rows <= 0 )
	    rows = self->arraysize ;


	/* Describe output SQLDA */

	isc_dsql_describe( isc_status, &self->STMT_hdl, 1, conn->o_sqlda ) ;
	CHECK_IBERR2( isc_status, NULL ) ;


	/* Set the output SQLDA */

	set_outputsqlda( self, conn->o_sqlda ) ;


	/* Create the list object */

	if( ( list = PyList_New( 0 ) ) == NULL )
	    return( NULL ) ;


	/* Fetch the specified number of rows */

	status = 0 ;
	for( count = 0 ; count < rows && !status ; count ++ )
	{
	    tuple = fetch_onerow( self, &status ) ;
	    if( !status )
	    {
		if( PyList_Append( list, tuple ) )
		{
		    PyMem_DEL( list ) ;
		    return( NULL ) ;
		}
	    }
	    else if( tuple == NULL )
	    {
		PyMem_DEL( list ) ;
		return( NULL ) ;
	    }
	}

	return( list ) ;
}



/* Fetch all rows */

#ifdef __STDC__
static PyObject	*ibcurs_fetchall
(
    ibcursobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibcurs_fetchall( self, args )
    ibcursobject	*self ;
    PyObject		*args ;
#endif
{
	ibconnobject	*conn = self->conn_p ;
	long		 status ;
	PyObject	*tuple ;
	PyObject	*list ;



	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Check if connected database */

	CHECK_CONNDB( conn, NULL ) ;


	/* Check if open cursor */

	if( !self->open_f )
	{
	    set_pyerror( 0, ERR_NOOPENCUR ) ;
	    return( NULL ) ;
	}


	/* Check the arguments */

	if( !PyArg_ParseTuple( args, "" ) )
	    return( NULL ) ;


	/* Describe output SQLDA */

	isc_dsql_describe( isc_status, &self->STMT_hdl, 1, conn->o_sqlda ) ;
	CHECK_IBERR2( isc_status, NULL ) ;


	/* Set the output SQLDA */

	set_outputsqlda( self, conn->o_sqlda ) ;


	/* Create the list object */

	if( ( list = PyList_New( 0 ) ) == NULL )
	    return( NULL ) ;


	/* Fetch all rows */

	tuple = fetch_onerow( self, &status ) ;
	while( !status )
	{
	    if( PyList_Append( list, tuple ) )
	    {
		PyMem_DEL( list ) ;
		return( NULL ) ;
	    }
	    tuple = fetch_onerow( self, &status ) ;
	}

	if( tuple == NULL )
	{
	    PyMem_DEL( list ) ;
	    return( NULL ) ;
	}

	return( list ) ;
}



/* List of methods defined in cursor object */

static struct PyMethodDef ibcurs_methods[] =
{
{ "close",	( PyCFunction )ibcurs_close,		METH_VARARGS,	NULL },
{ "execute",	( PyCFunction )ibcurs_execute,		METH_VARARGS,	NULL },
{ "fetchone",	( PyCFunction )ibcurs_fetchone,		METH_VARARGS,	NULL },
{ "fetchmany",	( PyCFunction )ibcurs_fetchmany,	METH_VARARGS,	NULL },
{ "fetchall",	( PyCFunction )ibcurs_fetchall,		METH_VARARGS,	NULL },
{ NULL,		NULL }
};
