/*
 *                       ibdate.c
 *
 *   - The module access to InterBase for the Python.
 *     ( The functions for Date object )
 *
 *                               Copyright (C) 1997,1998 by RIOS Corporation
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * RIOS Corporation makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.  By use of this software the user agrees to
 * indemnify and hold harmless RIOS Corporation from any  claims or
 * liability for loss arising out of such use.
 */


/* Deallocate the Date object */

#ifdef __STDC__
static void	*ibdate_dealloc
(
    ibdateobject	*self
)
#else
static void	*ibdate_dealloc( self )
    ibdateobject	*self ;
#endif
{
	PyMem_DEL( self ) ;
	return ;
}



/* Get the attribute of Date object */

#ifdef __STDC__
static PyObject	*ibdate_getattr
(
    ibdateobject	*self,
    char		*attr_name
)
#else
static PyObject	*ibdate_getattr( self, attr_name )
    ibdateobject	*self ;
    char		*attr_name ;
#endif
{
	if( !strcmp( attr_name, DATEATTR_WDAY ) )
	    return( PyInt_FromLong( self->dtime.tm_wday ) ) ;

	return( Py_FindMethod( ibdate_methods,
			( PyObject * )self, attr_name ) ) ;
}



/* Compare two Date objects */

#ifdef __STDC__
static int	 ibdate_compare
(
    ibdateobject	*self,
    ibdateobject	*comp
)
#else
static int	 ibdate_compare( self, comp )
    ibdateobject	*self ;
    ibdateobject	*comp ;
#endif
{
	time_t		 std ;
	time_t		 cmp ;
	struct tm	 wrk ;


	wrk = self->dtime ;
	std = mktime( &wrk ) ;

	wrk = comp->dtime ;
	cmp = mktime( &wrk ) ;

	return( ( std < cmp ) ? -1 : ( std > cmp ) ? 1 : 0 ) ;
}



/* Print the Date object */

#ifdef __STDC__
static int	 ibdate_print
(
    ibdateobject	*self,
    FILE		*fp,
    int			 flags
)
#else
static int	 ibdate_print( self, fp, flags )
    ibdateobject	*self ;
    FILE		*fp ;
    int			 flags ;
#endif
{
	sprintf( tempbuff_S, "%d.%d.%d %02d:%02d:%02d",
		self->dtime.tm_year + 1900,
		self->dtime.tm_mon + 1,
		self->dtime.tm_mday,
		self->dtime.tm_hour,
		self->dtime.tm_min,
		self->dtime.tm_sec	) ;

	fputs( tempbuff_S, fp ) ;
	return( 0 ) ;
}



/* Increase the time of the Date object */

#ifdef __STDC__
static PyObject	*ibdate_concat
(
    ibdateobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibdate_concat( self, args )
    ibdateobject	*self ;
    PyObject		*args ;
#endif
{
	ibdateobject	*newobj ;
	struct tm	 tms ;
	struct tm	*tm_p ;
	long		 inc ;
	time_t		 tval ;


	/* Check the parameter */

	if( PyInt_Check( args ) )
	    inc = PyInt_AS_LONG( ( PyIntObject * )args ) ;
	else
	{
	    set_pyerror( 0, ERR_DATEGAINV ) ;
	    return( NULL ) ;
	}


	/* Increase the current time value */

	tms = self->dtime ;
	tval = mktime( &tms ) + inc ;
	tm_p = localtime( &tval ) ;
	tms = *tm_p ;

	if( check_validdate( &tms ) )
	    return( NULL ) ;

	newobj = new_date_object( &tms ) ;
	if( newobj == NULL )
	    return( NULL ) ;

	return( ( PyObject * )newobj ) ;
}

/************************************************************************
 *  The following functions are written for the unsupported operations	*
 ************************************************************************/

#ifdef __STDC__
static int	 ibdate_length
(
    ibdateobject	*self
)
#else
static int	 ibdate_length( self )
    ibdateobject	*self ;
#endif
{
	set_pyerror( 0, ERR_DATEOPRND ) ;
	return( STATUS_ABNORMAL ) ;
}

#ifdef __STDC__
static PyObject	*ibdate_repeat
(
    ibdateobject	*self,
    int			 value
)
#else
static PyObject	*ibdate_repeat( self, value )
    ibdateobject	*self ;
    int			 value ;
#endif
{
	set_pyerror( 0, ERR_DATEOPRND ) ;
	return( NULL ) ;
}

#ifdef __STDC__
static PyObject	*ibdate_slice
(
    ibdateobject	*self,
    int			 value1,
    int			 value2
)
#else
static PyObject	*ibdate_slice( self, value1, value2 )
    ibdateobject	*self ;
    int			 value1 ;
    int			 value2 ;
#endif
{
	set_pyerror( 0, ERR_DATEOPRND ) ;
	return( NULL ) ;
}

#ifdef __STDC__
static PyObject	*ibdate_asslice
(
    ibdateobject	*self,
    int			 value1,
    int			 value2,
    ibdateobject	*ass
)
#else
static PyObject	*ibdate_asslice( self, value1, value2, ass )
    ibdateobject	*self ;
    int			 value1 ;
    int			 value2 ;
    ibdateobject	*ass ;
#endif
{
	set_pyerror( 0, ERR_DATEOPRND ) ;
	return( NULL ) ;
}

/************************************************************************
 *	End of the functions for the unsupported operations		*
 ************************************************************************/



/* Get the item of Date object */

#ifdef __STDC__
static PyObject	*ibdate_getitem
(
    ibdateobject	*self,
    int			 itemno
)
#else
static PyObject	*ibdate_getitem( self, itemno )
    ibdateobject	*self ;
    int			 itemno ;
#endif
{
	long		 value ;


	/* Get the specified item value */

	switch( itemno )
	{
	    case DATEITM_YEAR :
		value = self->dtime.tm_year + 1900 ;
		break ;
	    case DATEITM_MONTH :
		value = self->dtime.tm_mon + 1 ;
		break ;
	    case DATEITM_DAY :
		value = self->dtime.tm_mday ;
		break ;
	    case DATEITM_HOUR :
		value = self->dtime.tm_hour ;
		break ;
	    case DATEITM_MIN :
		value = self->dtime.tm_min ;
		break ;
	    case DATEITM_SEC :
		value = self->dtime.tm_sec ;
		break ;
	    default :
		set_pyerror( 0, ERR_DATEITMNO ) ;
		return( NULL ) ;
	}

	return( PyInt_FromLong( value ) ) ;
}



/* Get all items of date object */

#ifdef __STDC__
static PyObject	*ibdate_getitems
(
    ibdateobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibdate_getitems( self, args )
    ibdateobject	*self ;
    PyObject		*args ;
#endif
{
	PyObject	*tuple ;


	/* Check the arguments */

	if( !PyArg_ParseTuple( args, "" ) )
	    return( NULL ) ;


	/* Create the result tuple object */

	tuple = date_tuple_from_tm( &(self->dtime) ) ;

	return( tuple ) ;
}



/* Set the item of Date object */

#ifdef __STDC__
static int	ibdate_setitem
(
    ibdateobject	*self,
    int			 itemno,
    PyObject		*args
)
#else
static int	ibdate_setitem( self, itemno, args )
    ibdateobject	*self ;
    int			 itemno ;
    PyObject		*args ;
#endif
{
	long		 value ;
	struct tm	 tms ;


	/* Get the arguments */

	if( !PyInt_Check( args ) )
	{
	    set_pyerror( 0, ERR_DATEITMTP ) ;
	    return( STATUS_ABNORMAL ) ;
	}
	value = PyInt_AS_LONG( ( PyIntObject * )args ) ;


	/* Get the current values */

	tms = self->dtime ;


	/* Set the specified item value */

	switch( itemno )
	{
	    case DATEITM_YEAR :
		tms.tm_year = value - 1900 ;
		break ;
	    case DATEITM_MONTH :
		tms.tm_mon = value - 1 ;
		break ;
	    case DATEITM_DAY :
		tms.tm_mday = value ;
		break ;
	    case DATEITM_HOUR :
		tms.tm_hour = value ;
		break ;
	    case DATEITM_MIN :
		tms.tm_min = value ;
		break ;
	    case DATEITM_SEC :
		tms.tm_sec = value ;
		break ;
	    default :
		set_pyerror( 0, ERR_DATEITMNO ) ;
		return( STATUS_ABNORMAL ) ;
	}


	/* Check the date */

	if( check_validdate( &tms ) )
	    return( STATUS_ABNORMAL ) ;


	/* Set the new values */

	self->dtime = tms ;


	return( STATUS_NORMAL ) ;
}



/* Set all items of date object */

#ifdef __STDC__
static PyObject	*ibdate_setitems
(
    ibdateobject	*self,
    PyObject		*args
)
#else
static PyObject	*ibdate_setitems( self, args )
    ibdateobject	*self ;
    PyObject		*args ;
#endif
{
	PyObject	*params ;
	struct tm	 tms ;


	/* Get the arguments */

	if( !PyArg_ParseTuple( args, "O", &params ) )
	    return( NULL ) ;


	/* Set the date items into tm structure */

	if( tm_from_date_tuple( params, &tms ) )
	    return( NULL ) ;


	/* Set the new values */

	self->dtime = tms ;


	Py_INCREF( Py_None ) ;
	return( Py_None ) ;
}



/* List of methods defined in date object */

static struct PyMethodDef ibdate_methods[] =
{
{ "getitems",	( PyCFunction )ibdate_getitems,	METH_VARARGS,	NULL },
{ "setitems",	( PyCFunction )ibdate_setitems,	METH_VARARGS,	NULL },
{ NULL,		NULL }
};
