/*
 *                       ibmodule.c
 *
 *   - The module access to InterBase for the Python.
 *
 *                               Copyright (C) 1997,1998 by RIOS Corporation
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * RIOS Corporation makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.  By use of this software the user agrees to
 * indemnify and hold harmless RIOS Corporation from any  claims or
 * liability for loss arising out of such use.
 */

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <ibase.h>
#include "Python.h"
#include "ibmodule.h"

#ifdef linux
#include <math.h>
#endif

#ifdef __FreeBSD__
#include <float.h>
#endif


/* Version definition */

static char	*IB_module_version = IB_MODULE_VERSION ;


/* Database definitions */

static long		isc_status[20] ;
static long		SQLCODE ;
static char		isc_tpb_0[] =
    {
	isc_tpb_version1,	isc_tpb_write,
	isc_tpb_concurrency,	isc_tpb_nowait
    } ;
static char		isc_info_stmt[] =
    {	isc_info_sql_stmt_type	} ;
static char		isc_info_buff[16] ;


/* Variable definitions */

static PyObject	*ErrorObject ;
static PyObject	*VersionObject ;
static char	 tempbuff_S[BUFSIZ] ;
static long	 curs_num_S ;


/* Method list and object definitions */

extern struct PyMethodDef	ibcurs_methods[] ;
extern struct PyMethodDef	ibconn_methods[] ;
extern struct PyMethodDef	ibdate_methods[] ;
extern struct PyMethodDef	ibblob_methods[] ;
extern struct PyMethodDef	ibmdl_methods[] ;
extern PyTypeObject		IBcurstype ;
extern PyTypeObject		IBconntype ;
extern PyTypeObject		IBdatetype ;
extern PyTypeObject		IBblobtype ;



/************************************************************************
 *									*
 *	Common functions						*
 *									*
 ************************************************************************/

#include "ibcommon.c"


/************************************************************************
 *									*
 *	Functions for the transaction control				*
 *									*
 ************************************************************************/

#include "ibtrans.c"


/************************************************************************
 *									*
 *	Functions for the cursor object					*
 *									*
 ************************************************************************/

#include "ibcursor.c"


/************************************************************************
 *									*
 *	Functions for the blob object					*
 *									*
 ************************************************************************/

#include "ibblob.c"


/************************************************************************
 *									*
 *	Functions for the connection object				*
 *									*
 ************************************************************************/

#include "ibconn.c"


/************************************************************************
 *									*
 *	Functions for the date object					*
 *									*
 ************************************************************************/

#include "ibdate.c"


/************************************************************************
 *									*
 *	Functions for the module methods				*
 *									*
 ************************************************************************/

/* Connect to the database */

static keywd_args	Args_connect[] =
{
	{	"charset",	NULL	},
	{	"username",	NULL	},
	{	"password",	NULL	},
	{	NULL,		NULL	}
} ;

#ifdef __STDC__
static PyObject	*ibmdl_connect
(
    PyObject	*self,
    PyObject	*args,
    PyObject	*kwd
)
#else
static PyObject	*ibmdl_connect( self, args, kwd )
    PyObject	*self ;
    PyObject	*args ;
    PyObject	*kwd ;
#endif
{
	long		 count ;
	long		 DB_num ;
	DBinformation	*DB_info ;
	PyObject	*newobj ;
	char		*dbname ;



	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Get the arguments */

	for( count = 0 ; Args_connect[count].keyword ; count ++ )
	    Args_connect[count].value = NULL ;

	if( !PyArg_ParseTuple( args, "s|sss",
		&dbname,		&Args_connect[0].value,
		&Args_connect[1].value,	&Args_connect[2].value ) )
	    return( NULL ) ;

	if( !dbname || strlen( dbname ) == 0 )
	{
	    set_pyerror( 0, ERR_NODBNAME ) ;
	    return( NULL ) ;
	}

	if( get_keywordargs( kwd, Args_connect ) )
	    return( NULL ) ;


	/* Allocate the DB information structure */

	DB_num = 1 ;
	DB_info = ( DBinformation * )malloc( sizeof( DBinformation ) ) ;
	if( DB_info == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    return( NULL ) ;
	}


	/* Connect to the database */

	if( connect_database( DB_info, dbname, Args_connect[0].value,
			Args_connect[1].value, Args_connect[2].value, NULL ) )
	{
	    disconnect_database( DB_num, DB_info ) ;
	    free( DB_info ) ;
	    return( NULL ) ;
	}


	/* Create the connection object */

	newobj = create_connection( DB_num, DB_info ) ;
	if( newobj == NULL )
	{
	    disconnect_database( DB_num, DB_info ) ;
	    free( DB_info ) ;
	    return( NULL ) ;
	}


	return( newobj ) ;
}



/* Connect to the multiple database */

#ifdef __STDC__
static PyObject	*ibmdl_mconnect
(
    PyObject	*self,
    PyObject	*args
)
#else
static PyObject	*ibmdl_mconnect( self, args )
    PyObject	*self ;
    PyObject	*args ;
#endif
{
	PyObject	*param ;
	PyObject	*DB_list ;
	long		 DB_num ;
	long		 count ;
	PyObject	*( * func_getlist )() ;
	PyObject	*obj ;
	DBinformation	*DB_info ;
	PyObject	*( * func_getitem )() ;
	long		 item_num ;
	long		 icount ;

	PyObject	*newobj ;
	char		*dbname ;
	char		*alias ;
	char		*chset ;
	char		*usname ;
	char		*passwd ;



	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Get the arguments */

	if( !PyArg_ParseTuple( args, "O", &param ) )
	    return( NULL ) ;


	/* Check the arguments */

	if( PyTuple_Check( param ) && PyTuple_Size( param ) > 0 )
	{
	    func_getlist = PyTuple_GetItem ;
	    obj = PyTuple_GetItem( param, 0 ) ;
	    DB_num = PyTuple_Size( param ) ;
	}
	else if( PyList_Check( param ) && PyList_Size( param ) > 0 )
	{
	    func_getlist = PyList_GetItem ;
	    obj = PyList_GetItem( param, 0 ) ;
	    DB_num = PyList_Size( param ) ;
	}
	else
	{
	    set_pyerror( 0, ERR_MCONPARAM ) ;
	    return( NULL ) ;
	}

	if( PyString_Check( obj ) )
	{
	    func_getlist = NULL ;
	    DB_list = param ;
	    DB_num = 1 ;
	}
	else if( !PyTuple_Check( obj ) && !PyList_Check( obj ) )
	{
	    set_pyerror( 0, ERR_MCONPARAM ) ;
	    return( NULL ) ;
	}


	/* Allocate the DB information structure */

	DB_info = ( DBinformation * )malloc(
				sizeof( DBinformation ) * DB_num ) ;
	if( DB_info == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    return( NULL ) ;
	}


	/* Get the DB information and connect DB */

	for( count = 0 ; count < DB_num ; count ++ )
	{
	    if( func_getlist )
		DB_list = func_getlist( param, count ) ;


	    /* Get the parameters */

	    if( PyTuple_Check( DB_list ) )
	    {
		func_getitem = PyTuple_GetItem ;
		item_num = PyTuple_Size( DB_list ) ;
	    }
	    else if( PyList_Check( DB_list ) )
	    {
		func_getitem = PyList_GetItem ;
		item_num = PyList_Size( DB_list ) ;
	    }
	    else
	    {
		func_getitem = NULL ;
	    }

	    if( func_getitem == NULL
	    ||	item_num < 1 || item_num > CONN_PARM_MAX )
	    {
		set_pyerror( 0, ERR_MCONPARAM ) ;
		disconnect_database( count, DB_info ) ;
		free( DB_info ) ;
		return( NULL ) ;
	    }

	    dbname = chset = alias = usname = passwd = NULL ;
	    for( icount = 0 ; icount < item_num ; icount ++ )
	    {
		obj = func_getitem( DB_list, icount ) ;
		if( !PyString_Check( obj ) )
		{
		    set_pyerror( 0, ERR_MCONPARAM ) ;
		    disconnect_database( count, DB_info ) ;
		    free( DB_info ) ;
		    return( NULL ) ;
		}

		switch( icount )
		{
		    case 0 :
			dbname = PyString_AS_STRING( ( PyStringObject * )obj ) ;
			break ;
		    case 1 :
			alias = PyString_AS_STRING( ( PyStringObject * )obj ) ;
			break ;
		    case 2 :
			chset = PyString_AS_STRING( ( PyStringObject * )obj ) ;
			break ;
		    case 3 :
			usname = PyString_AS_STRING( ( PyStringObject * )obj ) ;
			break ;
		    case 4 :
			passwd = PyString_AS_STRING( ( PyStringObject * )obj ) ;
			break ;
		}
	    }


	    /* Check the database name */

	    if( !dbname || strlen( dbname ) == 0 )
	    {
		set_pyerror( 0, ERR_NODBNAME ) ;
		disconnect_database( count, DB_info ) ;
		free( DB_info ) ;
		return( NULL ) ;
	    }


	    /* Connect to the database */

	    if( connect_database( &DB_info[count], dbname,
					chset, usname, passwd, alias ) )
	    {
		disconnect_database( count, DB_info ) ;
		free( DB_info ) ;
		return( NULL ) ;
	    }
	}


	/* Create the connection object */

	newobj = create_connection( DB_num, DB_info ) ;
	if( newobj == NULL )
	{
	    disconnect_database( DB_num, DB_info ) ;
	    free( DB_info ) ;
	    return( NULL ) ;
	}


	return( newobj ) ;
}



#ifdef __STDC__
static PyObject	*ibmdl_date
(
    PyObject	*self,
    PyObject	*args
)
#else
static PyObject	*ibmdl_date( self, args )
    PyObject	*self ;
    PyObject	*args ;
#endif
{
	ibdateobject	*newobj ;
	PyObject	*params ;
	struct tm	 tms ;



	/* Clear the SQLCODE */

	CLEAR_IBERR ;


	/* Get the arguments */

	if( !PyArg_ParseTuple( args, "O", &params ) )
	    return( NULL ) ;


	/* Set the date items into tm structure */

	if( tm_from_date_tuple( params, &tms ) )
	    return( NULL ) ;


	/* Create the date object */

	newobj = new_date_object( &tms ) ;
	if( newobj == NULL )
	    return( NULL ) ;


	return( ( PyObject * )newobj ) ;
}



/* IB connection type object definition */

static PyTypeObject	IBconntype =
{
	PyObject_HEAD_INIT( &PyType_Type )
	0,				/* ob_size */
	"ibconnection",			/* tp_name */
	sizeof(ibconnobject),		/* tp_basicsize */
	0,				/* tp_itemsize */
	( destructor )ibconn_dealloc,	/* tp_dealloc */
	( printfunc )0,			/* tp_print */
	( getattrfunc )ibconn_getattr,	/* tp_getattr */
	( setattrfunc )0,		/* tp_setattr */
	( cmpfunc )0,			/* tp_compare */
	( reprfunc )0,			/* tp_repr */
	0,				/* tp_as_number */
	0,				/* tp_as_sequence */
	0,				/* tp_as_mapping */
	( hashfunc )0,			/* tp_hash */
	( ternaryfunc )0,		/* tp_call */
	( reprfunc )0,			/* tp_str */
	( getattrofunc )0,		/* tp_getattro */
	( setattrofunc )0,		/* tp_setattro */
	0,				/* tp_xxx3 */
	0,				/* tp_xxx4 */
	0				/* tp_doc */
} ;



/* IB date type object definition */

static PySequenceMethods	IBdate_seq =
{
	( inquiry )ibdate_length,		/* sq_length */
	( binaryfunc )ibdate_concat,		/* sq_concat */
	( intargfunc )ibdate_repeat,		/* sq_repeat */
	( intargfunc )ibdate_getitem,		/* sq_item */
	( intintargfunc )ibdate_slice,		/* sq_slice */
	( intobjargproc )ibdate_setitem,	/* sq_ass_item */
	( intintobjargproc )ibdate_asslice	/* sq_ass_slice */
};

static PyTypeObject	IBdatetype =
{
	PyObject_HEAD_INIT( &PyType_Type )
	0,				/* ob_size */
	"ibdatetype",			/* tp_name */
	sizeof(ibdateobject),		/* tp_basicsize */
	0,				/* tp_itemsize */
	( destructor )ibdate_dealloc,	/* tp_dealloc */
	( printfunc )ibdate_print,	/* tp_print */
	( getattrfunc )ibdate_getattr,	/* tp_getattr */
	( setattrfunc )0,		/* tp_setattr */
	( cmpfunc )ibdate_compare,	/* tp_compare */
	( reprfunc )0,			/* tp_repr */
	0,				/* tp_as_number */
	&IBdate_seq,			/* tp_as_sequence */
	0,				/* tp_as_mapping */
	( hashfunc )0,			/* tp_hash */
	( ternaryfunc )0,		/* tp_call */
	( reprfunc )0,			/* tp_str */
	( getattrofunc )0,		/* tp_getattro */
	( setattrofunc )0,		/* tp_setattro */
	0,				/* tp_xxx3 */
	0,				/* tp_xxx4 */
	0				/* tp_doc */
} ;



/* List of methods defined in the module */

static struct PyMethodDef ibmdl_methods[] =
{
{ "Connection",	( PyCFunction )ibmdl_connect,	METH_VARARGS|METH_KEYWORDS },
{ "Mconnection",( PyCFunction )ibmdl_mconnect,	METH_VARARGS },
{ "Date",	( PyCFunction )ibmdl_date,	METH_VARARGS },
{ NULL,		NULL }
};



/* Initialize function */

#ifdef __STDC__
void	initinterbasedb( void )
#else
void	initinterbasedb()
#endif
{
	PyObject	*mdl ;
	PyObject	*dic ;



	/* Create the module and add the functions */

	mdl = Py_InitModule( "interbasedb", ibmdl_methods ) ;


	/* Add some symbolic constants to the module */

	dic = PyModule_GetDict( mdl ) ;
	ErrorObject = Py_BuildValue( "s", "interbasedb.error" ) ;
	PyDict_SetItemString( dic, "error", ErrorObject ) ;

	VersionObject = Py_BuildValue( "s", IB_module_version ) ;
	PyDict_SetItemString( dic, "version", VersionObject ) ;

	if( PyErr_Occurred() )
	    Py_FatalError( "can't initialize module interbasedb" ) ;

	curs_num_S = 0 ;

	return ;
}
