/*
 *                       ibmodule.h
 *
 *   - The module access to InterBase for the Python.
 *
 *                               Copyright (C) 1997,1998 by RIOS Corporation
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * RIOS Corporation makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.  By use of this software the user agrees to
 * indemnify and hold harmless RIOS Corporation from any  claims or
 * liability for loss arising out of such use.
 */


#ifndef _IBMODULE_H_
#define _IBMODULE_H_


#include <time.h>



/* Version of interbasedb */

#define	IB_MODULE_VERSION	"Version 1.1 Revision 8"


/* Utility Macro definitions */

#define	ALIGN( n, b )	( ( n + b - 1 ) & ~( b - 1 ) )
#define	UPPER(c)	(((c) >= 'a' && (c)<= 'z') ? (c) - 'a' + 'A' : (c))
#define	FREE(p)		if( p )	{ free( p ) ; p = NULL ; }
#define	SETNULL(p)	if( p && strlen( p ) == 0 )	{ p = NULL ; }

#define	CLEAR_IBERR	SQLCODE=0

#define	CHECK_IBERR1(s)	SQLCODE=isc_sqlcode(s) ; if(SQLCODE) goto Error_Proc
#define	CHECK_IBERR2(s,r)	\
			SQLCODE=isc_sqlcode(s) ; if(SQLCODE)\
			{ set_dberror(s) ; return( (r) ) ; }

#define	CHECK_CONNDB(c,r)	\
			if( !(c) || !(c)->open_f || (c)->DB_num == 0 )\
			{	PyErr_SetString( ErrorObject, ERR_NODATABAS ) ;\
				return( (r) ) ;	}

#define	IS_DATEOBJECT(v)	( (v)->ob_type == &IBdatetype )
#define	IS_BLOBOBJECT(v)	( (v)->ob_type == &IBblobtype )


/* Return status code */

#define	STATUS_NORMAL	 0
#define	STATUS_ABNORMAL	-1


/* Statement type */

#define	STATEMENT_DDL	1
#define	STATEMENT_DML	0


/* Cursor open flag */

#define	IBOBJ_INIT	0
#define	IBOBJ_OPEN	1


/* Execute process flag */

#define	EXECF_EXECDML	0
#define	EXECF_SETPARM	1


/* Item No. of date object */

#define	DATEITM_YEAR	0
#define	DATEITM_MONTH	1
#define	DATEITM_DAY	2
#define	DATEITM_HOUR	3
#define	DATEITM_MIN	4
#define	DATEITM_SEC	5


/* Attribute name definitions */

#define	CURATTR_ARRAY	"arraysize"
#define	CURATTR_DESCS	"description"

#define	DATEATTR_WDAY	"dayofweek"

#define	BLOBATTR_TABLE	"table"
#define	BLOBATTR_FIELD	"field"
#define	BLOBATTR_SEGNO	"segment"
#define	BLOBATTR_SUBTP	"subtype"
#define	BLOBATTR_CHSET	"charset"
#define	BLOBATTR_DSIZE	"size"


/* Parameter definitions */

#define	RESULT_BUFINIT	BUFSIZ
#define	PARAMT_BUFINIT	BUFSIZ
#define	SQLDA_COLSINIT	10
#define	CURSOR_NAMEFMT	"c%d "
#define	COLATTR_NUMBER	7
#define	SQLCODE_NOMORE	100
#define	DATEOBJ_NUMBER	6
#define	BLOBINFO_RSLT	30
#define	USHORT_MAX	65535
#define	META_NAME_MAX	31
#define	BLOB_SBTP_TEXT	1
#define	CONN_PARM_MAX	5
#define	TPBBUFF_ALLOC	64
#define	CMND_DELIMIT	" \t\n\r\f"
#define	LIST_DELIMIT	", \t\n\r\f"


/* Character set name definitions */

#define	IBCHARSET_EUC	"EUCJ_0208"
#define	IBCHARSET_SJIS	"SJIS_0208"


/* Python object type definitions */

#define	PYOTP_STRING	1
#define	PYOTP_INTEGER	2
#define	PYOTP_FLOAT	3
#define	PYOTP_IBDATE	4
#define	PYOTP_IBBLOB	5
#define	PYOTP_OTHERS	0
#define	PYOTP_NONEOBJ	-1


/* DB handle and TR parameter block list structure */

typedef struct
{
	isc_db_handle	*dbb_ptr ;
	long		 tpb_len ;
	char		*tpb_ptr ;
} ISC_TEB ;


/* Database information structure definition */

typedef struct DBInformation
{
	char			*db_alias ;	/* Alias name */
	isc_db_handle		 DB_hdl ;	/* DB handle */
	long			 tpb_len ;	/* TR parameter block size */
	char			*tpb_ptr ;	/* TR parameter block */
	long			 charset ;	/* Character set */
	char			*username ;	/* User name */
	char			*password ;	/* Password */
} DBinformation ;


/* IB connection object definition */

typedef struct ConnectionObj
{
	PyObject_HEAD			/* Python object common header */
	long			 DB_num ;
	DBinformation		*DB_info ;
	isc_tr_handle		 TR_hdl ;
	long			 teb_size ;
	ISC_TEB			*teb_vec ;
	XSQLDA			*i_sqlda ;
	XSQLDA			*o_sqlda ;
	char			*results ;
	long			 ressize ;
	char			*paramts ;
	long			 prmsize ;
	long			 open_f ;
	struct CursorObj	*first_p ;
} ibconnobject ;


/* IB cursor object definition */

typedef struct CursorObj
{
	PyObject_HEAD			/* Python object common header */
	struct ConnectionObj	*conn_p ;
	char			*curs_name ;
	long			 DB_no ;
	isc_stmt_handle		 STMT_hdl ;
	long			 open_f ;
	long			 arraysize ;
	PyObject		*describe ;
	struct CursorObj	*prev_p ;
	struct CursorObj	*next_p ;
} ibcursobject ;


/* IB date object definition */

typedef struct DateObj
{
	PyObject_HEAD			/* Python object common header */
	struct tm		 dtime ;
} ibdateobject ;


/* IB blob object definition */

typedef struct BlobObj
{
	PyObject_HEAD			/* Python object common header */
	ISC_BLOB_DESC		 desc ;
	long			 size ;
	char			*blob ;
} ibblobobject ;


/* InterBase varchar structure */

typedef	struct
{
	short		 vary_length ;
	char		 vary_string[1];
} VARY ;


/* Transaction option list structure */

typedef struct trans_opts
{
	char			*option1 ;
	char			*option2 ;
	char			 optval ;
	short			 position ;
	struct trans_opts	*sub_opts ;
} trans_opts ;


/* Argument list structure */

typedef struct
{
	char			*keyword ;
	char			*value ;
} keywd_args ;



/* Messages */

#define	ERR_ALLOCMEM	"Failed to allocate memory."
#define	ERR_CONNPARAM	"Parameters must be string objects."
#define	ERR_MCONPARAM	"Parameter must be a tuple or list of the tuple or list of string objects."
#define	ERR_ILGLKWDIC	"Keyword parameter must be dictionary."
#define	ERR_NOTSTRING	"Keyword parameters must be string objects."
#define	ERR_DUPKWDARG	"Keyword parameter redefined."
#define	ERR_ILGLKWDPM	"Unknown Keyword : %s"
#define	ERR_NODBNAME	"Illegal database name was specified."
#define	ERR_NOALIASDB	"Specified database has not been connected."
#define	ERR_NODATABAS	"No database has been connected."
#define	ERR_TRSTARTED	"The transaction has been already started."
#define	ERR_NOOPENCUR	"The cursor has not been open.  Use execute(query)."
#define	ERR_ARRAYSIZE	"Cursor : arraysize must be greater than 0."
#define	ERR_ARRAYSINT	"Cursor : arraysize must be specified as int."
#define	ERR_NOINPARAM	"Input parameters must be specified."
#define	ERR_PARAMTYPE	"Parameters must be a tuple or list."
#define	ERR_UNMATCHNO	"Unmatch the number of parameters. Statement require %d items in each tuple or list."
#define	ERR_PARAMITEM	"Conversion error from parameter type to DB data type."
#define	ERR_UNSPPORTD	"Specified table includes unsupport datatype."
#define	ERR_STRTOINT	"Could not convert string to integer."
#define	ERR_SHORTOVER	"Short integer overflow."
#define	ERR_STRTOREAL	"Could not convert string to real."
#define	ERR_FLOATOVER	"Float overflow."
#define	ERR_SETNONULL	"Specified column is not permitted to set null."
#define	ERR_ILLGLDATE	"Specified date is invalid."
#define	ERR_DATEITMNO	"Item No. of date object must be pecified 0-5."
#define	ERR_DATEOTYPE	"The items of date must be specified tuple or list of int."
#define	ERR_NUMITEMDT	"The number of items for date object must be 1-6."
#define	ERR_DATEITMTP	"The item of date must be specified as int."
#define	ERR_DATEOPRND	"Date : bad operand type."
#define	ERR_DATEGAINV	"Date : specify the time[sec.] value of increase as int."

#define	ERR_USETRANS	"Use the method transaction() of Connection object."
#define	ERR_USECOMMIT	"Use the method commit() of Connection object."
#define	ERR_USEROLLBK	"Use the method rollback() of Connection object."

#define	ERR_ILGLTROPT	"Illegal transaction option was specified."
#define	ERR_DUPLTROPT	"Duplicate transaction option was specified."
#define	ERR_RESVNOTBL	"RESERVING needs table name list."
#define	ERR_ILGLTBLNM	"Illegal table name was specified."
#define	ERR_RESVUSAGE	"RESERVING needs {SHARED|PROTECTED} {READ|WRITE}."
#define	ERR_UNEXPCEOC	"Unexpected end of command."

#define	ERR_BLOBOBJCT	"Blob object need table name and field name."
#define	ERR_BLOBDATTP	"Blob data must be specified as String."

#define	ERR_NOSPARRAY	"Not support the operation to the array field."



#endif
