/*
 *                       ibtrans.c
 *
 *   - The module access to InterBase for the Python.
 *     ( The functions for control transaction )
 *
 *                               Copyright (C) 1997,1998 by RIOS Corporation
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * RIOS Corporation makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.  By use of this software the user agrees to
 * indemnify and hold harmless RIOS Corporation from any  claims or
 * liability for loss arising out of such use.
 */


/* Transaction option list */

static trans_opts	rcom_opt_S[] =
{
"NO",	"RECORD_VERSION",	isc_tpb_no_rec_version,	-1,	NULL,
"RECORD_VERSION",	NULL,	isc_tpb_rec_version,	-1,	NULL,
"*",		NULL,		isc_tpb_no_rec_version,	-1,	NULL,
0,		NULL,		0,			0,	NULL
} ;


static trans_opts	read_opt_S[] =
{
"WRITE",	NULL,		isc_tpb_write,		1,	NULL,
"ONLY",		NULL,		isc_tpb_read,		1,	NULL,
"COMMITTED",	NULL,		isc_tpb_read_committed,	2,	rcom_opt_S,
0,		NULL,		0,			0,	NULL
} ;


static trans_opts	snap_opt_S[] =
{
"TABLE",	"STABILITY",	isc_tpb_consistency,	2,	NULL,
"*",		NULL,		isc_tpb_concurrency,	2,	NULL,
0,		NULL,		0,			0,	NULL
} ;


static trans_opts	isol_opt_S[] =
{
"SNAPSHOT",	NULL,		0,			0,	snap_opt_S,
"READ",		"COMMITTED",	isc_tpb_read_committed,	2,	rcom_opt_S,
0,		NULL,		0,			0,	NULL
} ;


static trans_opts	trans_opt_S[] =
{
"READ",		NULL,		0,			0,	read_opt_S,
"WAIT",		NULL,		isc_tpb_wait,		3,	NULL,
"NO",		"WAIT",		isc_tpb_nowait,		3,	NULL,
"ISOLATION",	"LEVEL",	0,			0,	isol_opt_S,
"SNAPSHOT",	NULL,		0,			0,	snap_opt_S,
"RESERVING",	NULL,		-1,			0,	NULL,
0,		NULL,		0,			0,	NULL
} ;

/* Name1	Name2		Option value	    Position	Sub-option */

#define	RESV_TABLEEND	"FOR"
#define	RESV_SHARED	"SHARED"
#define	RESV_PROTECTD	"PROTECTED"
#define	RESV_READ	"READ"
#define	RESV_WRITE	"WRITE"
#define	RESV_CONTINUE	','




/* Start transaction */

#ifdef __STDC__
static int	 start_transaction
(
    ibconnobject	*conn
)
#else
static int	 start_transaction( conn )
    ibconnobject	*conn ;
#endif
{
	DBinformation	*DB_info ;
	long		 count ;



	/* Set the transaction options */

	if( conn->DB_num > conn->teb_size )
	{
	    if( conn->teb_vec )
		conn->teb_vec = ( ISC_TEB * )realloc(
					( void * )conn->teb_vec,
					sizeof( ISC_TEB ) * conn->DB_num ) ;
	    else
		conn->teb_vec = ( ISC_TEB * )malloc( sizeof( ISC_TEB )
							* conn->DB_num ) ;
	    if( conn->teb_vec == NULL )
	    {
		set_pyerror( 0, ERR_ALLOCMEM ) ;
		return( STATUS_ABNORMAL ) ;
	    }
	    conn->teb_size = conn->DB_num ;
	}

	for( count = 0 ; count < conn->DB_num ; count ++ )
	{
	    DB_info = &conn->DB_info[count] ;
	    conn->teb_vec[count].dbb_ptr = &DB_info->DB_hdl ;
	    if( DB_info->tpb_ptr )
	    {
		conn->teb_vec[count].tpb_len = DB_info->tpb_len ;
		conn->teb_vec[count].tpb_ptr = DB_info->tpb_ptr ;
	    }
	    else
	    {
		conn->teb_vec[count].tpb_len = sizeof( isc_tpb_0 ) ;
		conn->teb_vec[count].tpb_ptr = isc_tpb_0 ;
	    }
	}


	/* Start transaction */

	isc_start_multiple( isc_status, &conn->TR_hdl,
			( short )conn->DB_num, conn->teb_vec ) ;
	CHECK_IBERR1( isc_status ) ;


	return( STATUS_NORMAL ) ;


Error_Proc:

	set_dberror( isc_status ) ;
	return( STATUS_ABNORMAL ) ;
}



/* Parse transaction options */

#ifdef __STDC__
static int	parse_transopts
(
    char	 *trans,
    char	**tpb_ptr,
    long	 *tpb_len
)
#else
static int	parse_transopts( trans, tpb_ptr, tpb_len )
    char	 *trans ;
    char	**tpb_ptr ;
    long	 *tpb_len ;
#endif
{
	long		 used ;
	long		 size ;
	char		*tpb ;
	trans_opts	*curr_p ;
	trans_opts	*target_p ;
	char		*check1_p ;
	char		*check2_p ;
	int		 count ;
	int		 next_c ;
	char		 check_f[4] ;
	char		*resv_p ;
	char		*resend_p ;
	char		*tblend_p ;
	int		 tbl_len ;
	int		 res_first ;
	int		 res_count ;
	int		 ofs ;
	char		 sp_prm ;
	char		 rw_prm ;
	int		 cont_f ;



	/* Initialize */

	used = 0 ;
	size = 0 ;
	tpb = NULL ;
	memset( ( void * )check_f, 0, sizeof( check_f ) ) ;


	/* Set the default transaction option */

	tpb = ( char * )malloc( TPBBUFF_ALLOC ) ;
	if( tpb == NULL )
	{
	    set_pyerror( 0, ERR_ALLOCMEM ) ;
	    return( STATUS_ABNORMAL ) ;
	}
	size = TPBBUFF_ALLOC ;
	memcpy( ( void * )tpb, ( void * )isc_tpb_0, sizeof( isc_tpb_0 ) ) ;
	used = sizeof( isc_tpb_0 ) ;


	/* Analize the transaction option strings */

	curr_p = trans_opt_S ;
	check1_p = strtok( trans, CMND_DELIMIT ) ;
	if( check1_p )
	    check2_p = strtok( NULL, CMND_DELIMIT ) ;
	else
	    check2_p = NULL ;
	while( curr_p )
	{
	    target_p = NULL ;
	    next_c = 0 ;
	    for( count = 0 ; curr_p[count].option1 ; count ++ )
	    {
		if( !strcmp( curr_p[count].option1, "*" ) )
		{
		    target_p = &curr_p[count] ;
		    break ;
		}
		else if( check1_p && !strcmp( check1_p, curr_p[count].option1 ))
		{
		    if( !curr_p[count].option2 )
		    {
			next_c = 1 ;
			target_p = &curr_p[count] ;
			break ;
		    }
		    else if( check2_p
		    &&	!strcmp( check2_p, curr_p[count].option2 ) )
		    {
			next_c = 2 ;
			target_p = &curr_p[count] ;
			break ;
		    }
		}
	    }

	    if( !target_p )
	    {
		free( tpb ) ;
		set_pyerror( 0, ERR_ILGLTROPT ) ;
		return( STATUS_ABNORMAL ) ;
	    }


	    /* Set the transaction option */

	    if( target_p->optval > '\0' )
	    {
		if( target_p->position > 0 )
		{
		    if( check_f[target_p->position] )
		    {
			free( tpb ) ;
			set_pyerror( 0, ERR_DUPLTROPT ) ;
			return( STATUS_ABNORMAL ) ;
		    }
		    tpb[target_p->position] = target_p->optval ;
		    check_f[target_p->position] = 1 ;
		}
		else
		{
		    if( used + 1 > size )
		    {
			tpb = ( char * )realloc( tpb, size + TPBBUFF_ALLOC ) ;
			if( tpb == NULL )
			{
			    free( tpb ) ;
			    set_pyerror( 0, ERR_ALLOCMEM ) ;
			    return( STATUS_ABNORMAL ) ;
			}
			size += TPBBUFF_ALLOC ;
		    }
		    tpb[used] = target_p->optval ;
		    used ++ ;
		}
	    }
	    else if( target_p->optval )		/* RESERVING ... FOR */
	    {
		if( check_f[0] )
		{
		    free( tpb ) ;
		    set_pyerror( 0, ERR_DUPLTROPT ) ;
		    return( STATUS_ABNORMAL ) ;
		}

		resv_p = check2_p ;
		if( !resv_p || !strcmp( resv_p, RESV_TABLEEND ) )
		{
		    free( tpb ) ;
		    set_pyerror( 0, ERR_RESVNOTBL ) ;
		    return( STATUS_ABNORMAL ) ;
		}

		while( resv_p )
		{
		    res_first = used ;
		    res_count = 0 ;
		    resend_p = strtok( NULL, CMND_DELIMIT ) ;
		    while( resend_p )
		    {
			if( !strcmp( resend_p, RESV_TABLEEND ) )
			    break ;
			resend_p = strtok( NULL, CMND_DELIMIT ) ;
		    }

		    if( !resend_p )
		    {
			free( tpb ) ;
			set_pyerror( 0, ERR_ILGLTROPT ) ;
			return( STATUS_ABNORMAL ) ;
		    }

		    while( resv_p < resend_p )
		    {
			if( *resv_p == '\0'
			|| ( ofs = strspn( resv_p, LIST_DELIMIT ) ) < 0 )
			    resv_p ++ ;
			else
			{
			    resv_p = &resv_p[ofs] ;
			    tblend_p = strpbrk( resv_p, LIST_DELIMIT ) ;
			    if( tblend_p )
				tbl_len = tblend_p - resv_p ;
			    else
				tbl_len = strlen( resv_p ) ;

			    if( tbl_len > META_NAME_MAX )
			    {
				free( tpb ) ;
				set_pyerror( 0, ERR_ILGLTBLNM ) ;
				return( STATUS_ABNORMAL ) ;
			    }

			    if( tbl_len > 0 )
			    {
				if( used + tbl_len + 3 > size )
				{
				    tpb = ( char * )realloc( tpb,
							size + TPBBUFF_ALLOC ) ;
				    if( tpb == NULL )
				    {
					free( tpb ) ;
					set_pyerror( 0, ERR_ALLOCMEM ) ;
					return( STATUS_ABNORMAL ) ;
				    }
				    size += TPBBUFF_ALLOC ;
				}
				tpb[used+1] = ( char )tbl_len ;
				memcpy( ( void * )&tpb[used+2],
							resv_p, tbl_len ) ;
				used += tbl_len + 3 ;
				res_count ++ ;
			    }
			    resv_p += tbl_len ;
			}
		    }

		    resv_p = strtok( NULL, CMND_DELIMIT ) ;
		    if( resv_p && !strcmp( resv_p, RESV_SHARED ) )
			sp_prm = isc_tpb_shared ;
		    else if( resv_p && !strcmp( resv_p, RESV_PROTECTD ) )
			sp_prm = isc_tpb_protected ;
		    else
		    {
			free( tpb ) ;
			set_pyerror( 0, ERR_RESVUSAGE ) ;
			return( STATUS_ABNORMAL ) ;
		    }

		    cont_f = 0 ;
		    resv_p = strtok( NULL, CMND_DELIMIT ) ;
		    if( resv_p )
		    {
			if( resv_p[strlen(resv_p)-1] == RESV_CONTINUE )
			{
			    cont_f = 1 ;
			    resv_p[strlen(resv_p)-1] = '\0' ;
			}
			else
			{
			    tblend_p = strpbrk( resv_p, LIST_DELIMIT ) ;
			    if( tblend_p )
			    {
				cont_f = 2 ;
				*tblend_p = '\0' ;
			    }
			}
		    }

		    if( resv_p && !strcmp( resv_p, RESV_READ ) )
			rw_prm = isc_tpb_lock_read ;
		    else if( resv_p && !strcmp( resv_p, RESV_WRITE ) )
			rw_prm = isc_tpb_lock_write ;
		    else
		    {
			free( tpb ) ;
			set_pyerror( 0, ERR_RESVUSAGE ) ;
			return( STATUS_ABNORMAL ) ;
		    }

		    ofs = res_first ;
		    for( count = 0 ; count < res_count ; count ++ )
		    {
			tpb[ofs++] = rw_prm ;
			ofs += tpb[ofs] + 1 ;
			tpb[ofs++] = sp_prm ;
		    }

		    if( cont_f == 1 )
		    {
			resv_p = strtok( NULL, CMND_DELIMIT ) ;
			if( !resv_p )
			{
			    free( tpb ) ;
			    set_pyerror( 0, ERR_UNEXPCEOC ) ;
			    return( STATUS_ABNORMAL ) ;
			}
		    }
		    if( cont_f == 2 )
		    {
			resv_p = tblend_p + 1 ;
		    }
		    else
		    {
			resv_p = strtok( NULL, CMND_DELIMIT ) ;
			if( resv_p )
			{
			    if( ( int )strlen( resv_p ) == 1
			    &&	resv_p[0] == RESV_CONTINUE )
			    {
				resv_p = strtok( NULL, CMND_DELIMIT ) ;
				if( !resv_p )
				{
				    free( tpb ) ;
				    set_pyerror( 0, ERR_UNEXPCEOC ) ;
				    return( STATUS_ABNORMAL ) ;
				}
			    }
			    else if( resv_p[0] == RESV_CONTINUE )
				resv_p ++ ;
			    else
			    {
				next_c = 1 ;
				check2_p = resv_p ;
				resv_p = NULL ;
			    }
			}
			else
			{
			    next_c = 0 ;
			    check1_p = check2_p = NULL ;
			}
		    }
		}

		check_f[0] = 1 ;
	    }


	    /* Set the next check list */

	    curr_p = target_p->sub_opts ;

	    for( count = 0 ; count < next_c ; count ++ )
	    {
		check1_p = check2_p ;
		if( check2_p )
		    check2_p = strtok( NULL, CMND_DELIMIT ) ;
	    }

	    if( check1_p && !curr_p )
		curr_p = trans_opt_S ;
	}


	/* Set the results */

	*tpb_ptr = tpb ;
	*tpb_len = used ;


	return( STATUS_NORMAL ) ;
}
